from typing import Optional
import gromologist as gml


class Subsection:
    counter = {}
    
    def __init__(self, content: list, section: "gml.Section"):
        """
        Here we want to have:
          - a unique representation of the section (header/ID)
          - a list of entries
          - a binding to the Top class that holds the Section object

        :param content: list of strings, entire content of the section
        :param section: a Section instance that contains this Subsection
        """
        self.section = section
        self.conditional = False
        self.header = content[0].strip().strip('[]').strip()
        if ';' in self.header:
            pos = self.header.index(';')
            self.header = self.header[:pos].strip().strip('[]').strip()
        if self.header in Subsection.counter.keys():
            Subsection.counter[self.header] += 1
        else:
            Subsection.counter[self.header] = 1
        self.id = Subsection.counter[self.header]
        self.entries = []
        for element in content:
            if issubclass(type(element), gml.Entry):
                self.entries.append(element)
            elif isinstance(element, str) and element.strip() and not element.strip().startswith('['):
                self.entries.append(self.yield_entry(element))
        
    def yield_entry(self, line: str) -> "gml.Entry":
        """
        Decides which Entry subclass to return
        based on which Subsection subclass evokes this fn
        :param line: str, a line to be converted into an Entry instance
        :return: Entry, an instance of the proper Entry subclass
        """
        if line.strip()[0] in [';', '#']:
            return gml.Entry(line, self)
        elif isinstance(self, SubsectionParam):
            return gml.EntryParam(line, self)
        elif isinstance(self, SubsectionBonded):
            return gml.EntryBonded(line, self)
        elif isinstance(self, SubsectionAtom):
            return gml.EntryAtom(line, self)
        elif isinstance(self, Subsection):
            return gml.Entry(line, self)
    
    def __str__(self) -> str:
        """
        As section headers can be repeated, each section is denoted
        by a header and ID (ID corresponding to the consecutive numbering
        of the specific header, e.g. bonds-3 is the third "bonds" section
        :return: str, section label
        """
        return "{}-{}".format(self.header, self.id)
    
    def __repr__(self) -> str:
        return "Subsection {}".format(self.header, self.id)
    
    def __len__(self) -> int:
        return len(self.entries)
    
    def __iter__(self):
        """
        Useful if we want to iterate over entries as "for entry in subsection",
        allows us to mark self._entries as private
        :return: self
        """
        self.n = 0
        return self
    
    def __next__(self):
        n = self.n
        self.n += 1
        try:
            return self.entries[n]
        except IndexError:
            raise StopIteration

    def add_entry(self, new_entry: "gml.Entry", position: Optional[int] = None):
        """
        Adds a single entry to the subsection, either at the end
        or in a specified position
        :param new_entry: gml.Entry, entry to be added
        :param position: where to add the entry (None is at the end)
        :return: None
        """
        new_entry.subsection = self
        if position is not None:
            position = int(position)
            self.entries.insert(position, new_entry)
        else:
            self.entries.append(new_entry)
    
    def add_entries(self, new_entries_list: list, position: Optional[int] = None):
        """
        Adds multiple entries to the subsection, either at the end
        or in a specified position
        :param new_entries_list: list of gml.Entry, entries to be added
        :param position: where to add the entries (None is at the end)
        :return: None
        """
        for new_entry in new_entries_list:
            new_entry.subsection = self
        if position is not None:
            position = int(position)
            for new_entry in new_entries_list:
                self.entries.insert(position, new_entry)
                position += 1
        else:
            self.entries.extend(new_entries_list)
    
    def set_entry(self, line_number: int, new_line: str):
        """
        Sets content of a specified entry
        :param line_number: int, which entry to modify
        :param new_line: str, new content of the entry
        :return: None
        """
        self.entries[line_number] = new_line
    
    def get_entry(self, line_number: int) -> "gml.Entry":
        """
        Returns entry specified by line number
        :param line_number: int, which entry to return
        :return: gml.Entry, subsection entry
        """
        return self.entries[line_number]

    def remove_entry(self, entry: "gml.Entry"):
        self.entries.remove(entry)
        
        
class SubsectionBonded(Subsection):
    """
    SubsectionBonded contains a subsection with entries corresponding to bonded terms,
    e.g., bonds or dihedrals; should be included in SectionMol
    """
    n_atoms = {'bonds': 2, 'pairs': 2, 'angles': 3, 'dihedrals': 4,
               'cmap': 5, 'settles': 1, 'exclusions': 2, 'position_restraints': 1,
               'virtual_sites2': 3, 'constraints': 2}
    
    def __init__(self, content: list, section: "gml.Section"):
        super().__init__(content, section)
        self.bkp_entries = None
        self.atoms_per_entry = SubsectionBonded.n_atoms[self.header]
        self.prmtypes = self._check_parm_type()
        self.fstring = "{:5} " * (SubsectionBonded.n_atoms[self.header] + 1) + '\n'
    
    def __repr__(self) -> str:
        return "Subsection {} with interaction type {}".format(self.header, ' '.join(self.prmtypes))

    @property
    def label(self) -> str:
        return '{}-{}'.format(self.header, '_'.join(self.prmtypes))

    @property
    def entries_bonded(self) -> list:
        """
        Lists all entries that are of EntryBonded type (i.e. excluding comments,
        headers, empty lines etc)
        :return: list of EntryBonded instances
        """
        return [e for e in self.entries if isinstance(e, gml.EntryBonded)]
    
    def sort(self):
        """
        In case we want to sort entries after some are added at the end of the section
        :return: None
        """
        self.entries.sort(key=self._sorting_fn)

    def change_interaction_type(self, change_from, change_to):
        """
        Changes interaction type of the whole section
        :param change_from: str/int, original interaction type
        :param change_to: str/int, target interaction type
        :return: None
        """
        change_from, change_to = str(change_from), str(change_to)
        for entry in self.entries_bonded:
            if entry.interaction_type == change_from:
                if entry.params_state_a:
                    self.section.top.print(f"Entry {str(entry)} has parameters explicitly assigned, skipping to "
                                           f"avoid errors")
                    continue
                entry.interaction_type = change_to

    def _sorting_fn(self, entry: "gml.EntryBonded") -> int:
        """
        Comments should go first, then we sort based on first, second,
        ... column of the section
        :param entry: Entry, entry to be sorted
        :return: int, ordering number
        """
        if isinstance(entry, gml.Entry):
            return -1
        val = sum([i * 10**(4*(self.atoms_per_entry - n)) for n, i in enumerate(entry.atom_numbers)])
        return val

    def explicit_defines(self):
        """
        Substitutes predefined fields with their corresponding values
        as specified by the #DEFINE preprocessor commands
        :return: None
        """
        for entry in self.entries_bonded:
            entry.explicit_defines()
    
    def add_ff_params(self, force_all: bool = False):
        """
        Adds explicit values of FF parameters in sections 'bonds', 'angles', 'dihedrals'
        :param force_all: bool, whether to add params from scratch if some are assigned already
        :return: None
        """
        matchings = {'bonds': 'bondtypes', 'angles': 'angletypes', 'dihedrals': 'dihedraltypes'}
        subsect_params = [sub for sub in self.section.top.parameters.subsections
                          if sub.header == matchings[self.header]]
        self.bkp_entries = self.entries[:]  # we can't change what we're iterating over, so we modify the copy
        for entry in self.entries_bonded:
            # let's omit the ones that already have params assigned:
            if entry.params_state_a:
                if not force_all:
                    continue
                else:
                    entry.params_state_a = []
                    entry.params_state_b = []
                    entry.params_state_a_entry = []
                    entry.params_state_b_entry = []
            self._add_ff_params_to_entry(entry, subsect_params)
        self.entries = self.bkp_entries[:]  # now restore the modified copy

    def find_used_ff_params(self) -> list:
        """
        Looks up a list of FF parameters that are actually used in the molecular
        system at hand
        :return: list of str, labels of used parameters
        """
        used_parm_entries = []
        matchings = {'bonds': 'bondtypes', 'angles': 'angletypes', 'dihedrals': 'dihedraltypes', 'cmap': 'cmaptypes'}
        subsect_params = [sub for sub in self.section.top.parameters.subsections if
                          sub.header == matchings[self.header]]
        for entry in self.entries_bonded:
            if not entry.params_state_a:
                used_parm_entries.extend(self._find_used_ff_params(entry, subsect_params))
        return used_parm_entries

    @staticmethod
    def _find_used_ff_params(entry: "gml.EntryBonded", subsect_params: list) -> list:
        """
        Identifies the FF parameter(s) that is used for the particular
        bonded interaction entry
        :param entry: an EntryBonded instance, an entry we're finding the parameter for
        :param subsect_params: a list of SubsectionParam instances containing candidate parameters for the entry
        :return: list of str, labels of used parameters
        """
        entries = []
        int_type = entry.interaction_type
        entry.read_types()
        for types in [entry.types_state_a, entry.types_state_b]:
            wildcard_present = []
            non_wildcard_present = []
            for subsections in subsect_params:
                for parm_entry in [e for e in subsections.entries_param]:
                    if parm_entry.match(types, int_type):
                        is_wildcard = 'X' in parm_entry.types
                        if not wildcard_present and not is_wildcard:
                            entries.append(parm_entry.identifier)
                            non_wildcard_present += parm_entry.types
                        elif not wildcard_present and is_wildcard and not non_wildcard_present:
                            entries.append(parm_entry.identifier)
                            wildcard_present = parm_entry.types
                        elif wildcard_present and not is_wildcard:
                            raise RuntimeError("Wildcard ('X') entries were found prior to regular ones, please fix"
                                               "your FF parameters")
                        elif wildcard_present and is_wildcard:  # only add if multiple entries per given wildcard
                            if parm_entry.types == wildcard_present:
                                entries.append(parm_entry.identifier)
                            else:
                                pass
        return entries

    def find_missing_ff_params(self, fix_by_analogy: bool = False, fix_B_from_A: bool = False,
                               fix_A_from_B: bool = False, fix_dummy: bool = False, once: bool = False):
        """
        Identifies FF parameters that cannot by matched to the existing set of
        bonded interactions, allowing to find & target these for fixing
        :param fix_by_analogy: dict, tells how to choose new parameters from existing equivalent types
        :param fix_B_from_A: bool, chooses missing parameters assuming B-state should have similar types to A-state
        :param fix_A_from_B: bool, chooses missing parameters assuming A-state should have similar types to B-state
        :param fix_dummy: bool, adds dummy (all-0) parameters
        :param once: bool, only shows the missing parameter once if repeated
        :return: None
        """
        matchings = {'bonds': 'bondtypes', 'angles': 'angletypes', 'dihedrals': 'dihedraltypes'}
        # TODO add cmaps?
        subsect_params = [sub for sub in self.section.top.parameters.subsections if
                          sub.header == matchings[self.header]]
        for entry in self.entries_bonded:
            self._find_missing_ff_params(entry, subsect_params, fix_by_analogy, fix_B_from_A, fix_A_from_B, fix_dummy,
                                         once)

    def _find_missing_ff_params(self, entry, subsect_params, fix_by_analogy, fix_B_from_A, fix_A_from_B, fix_dummy,
                                once):
        if (entry.params_state_a and entry.params_state_b) or (entry.params_state_a and not entry.types_state_b):
            return
        int_type = entry.interaction_type
        entry.read_types()
        found_a, found_b = False, False
        for subsections in subsect_params:
            for parm_entry in subsections.entries_param:
                if parm_entry.match(entry.types_state_a, int_type):
                    found_a = True
        if not found_a and not entry.params_state_a:
            if not once or entry.types_state_a not in self.section.printed:
                print(f'Couldn\'t find params for interaction type {entry.subsection.header} {int_type}, '
                      f'atom types {entry.types_state_a}, atom numbers {entry.atom_numbers}')
            if once and entry.types_state_a not in self.section.printed:
                self.section.printed.append(entry.types_state_a)
            if fix_by_analogy:
                candid = self._fix_by_analogy(fix_by_analogy, entry.types_state_a, subsect_params, int_type)
                if candid:
                    entry.params_state_a = candid
            if fix_dummy:
                entry.params_state_a = [0.0 for _ in gml.EntryBonded.fstr_suff[(self.header, self.prmtypes[0])]]
                print("setting dummy parameters: " + str(entry))
        if entry.types_state_b and not entry.params_state_b:
            for subsections in subsect_params:
                for parm_entry in subsections.entries_param:
                    if parm_entry.match(entry.types_state_b, int_type):
                        found_b = True
            if not found_b and not entry.params_state_b:
                if not once or entry.types_state_b not in self.section.printed:
                    print(f'Couldn\'t find params for interaction type {entry.subsection.header} {int_type}, '
                          f'atom types {entry.types_state_b}, atom numbers {entry.atom_numbers}')
                if once and entry.types_state_b not in self.section.printed:
                    self.section.printed.append(entry.types_state_b)
                if fix_by_analogy:
                    candid = self._fix_by_analogy(fix_by_analogy, entry.types_state_b, subsect_params, int_type)
                    if candid:
                        entry.params_state_b = candid
        if entry.types_state_b:
            if fix_B_from_A and not found_b and not entry.params_state_b:
                candid = self._fix_by_analogy({}, entry.types_state_b, subsect_params, int_type,
                                              other_typelist=entry.types_state_a)
                if candid:
                    entry.params_state_b = candid
            if fix_A_from_B and not found_a and not entry.params_state_a:
                candid = self._fix_by_analogy({}, entry.types_state_a, subsect_params, int_type,
                                              other_typelist=entry.types_state_b)
                if candid:
                    entry.params_state_a = candid

    @staticmethod
    def _fix_by_analogy(subst, typelist, subsect_params, int_type, other_typelist=None):
        new_params = []
        from_wildtype = None
        if other_typelist is None:
            types = [tp if tp not in subst.keys() else subst[tp] for tp in typelist]
        else:
            types = other_typelist
        for subsection in subsect_params:
            for parm_entry in subsection.entries_param:
                if parm_entry.match(types, int_type):
                    if from_wildtype is None:
                        if 'X' in parm_entry.types:
                            from_wildtype = True
                        else:
                            from_wildtype = False
                        new_params.append(parm_entry.params)
                        print(f'Fixing by analogy, using entry {str(parm_entry).strip()}')
                    elif (from_wildtype and 'X' in parm_entry.types) or (not from_wildtype
                                                                         and 'X' not in parm_entry.types):
                        new_params.append(parm_entry.params)
                        print(f'Fixing by analogy, using entry {str(parm_entry).strip()}')
                    if len(typelist) < 4:
                        break
            else:
                continue
            break
        return [y for x in new_params for y in x]
    
    def _add_ff_params_to_entry(self, entry, subsect_params):
        """
        Given a bonded term (e.g. "21     24     26    5") converts it to atomtypes,
        finds the respective FF parameters and adds them to the bonded entry
        :param entry: Entry, an EntryBonded instance to add FF params to
        :param subsect_params: list, SubsectionParam instances that hold all FF params
        :return: None
        """
        int_type = entry.interaction_type
        entry.read_types()
        for types, params, parmentry in zip([entry.types_state_a, entry.types_state_b],
                                            [entry.params_state_a, entry.params_state_b],
                                            [entry.params_state_a_entry, entry.params_state_b_entry]):
            wildcard_present = []
            non_wildcard_present = []
            for subsections in subsect_params:
                for parm_entry in subsections.entries_param:
                    if parm_entry.match(types, int_type):
                        is_wildcard = 'X' in parm_entry.types
                        if not wildcard_present and not is_wildcard:
                            params += parm_entry.params
                            parmentry.append(parm_entry)
                            non_wildcard_present += parm_entry.types
                        elif not wildcard_present and is_wildcard and not non_wildcard_present:
                            params += parm_entry.params
                            parmentry.append(parm_entry)
                            wildcard_present = parm_entry.types
                        elif wildcard_present and not is_wildcard:
                            raise RuntimeError("Wildcard ('X') entries were found prior to regular ones, please fix"
                                               "your FF parameters")
                        elif wildcard_present and is_wildcard:  # only add if multiple entries per given wildcard
                            if parm_entry.types == wildcard_present:
                                params += parm_entry.params
                                parmentry.append(parm_entry)
                            else:
                                pass
        if not entry.params_state_a and entry.subsection.header == 'dihedrals' \
                and any('DUM' in t for t in entry.types_state_a):
            entry.params_state_a = [0.0, 0.0, '1']
        if not entry.params_state_b and entry.subsection.header == 'dihedrals' and entry.types_state_b \
                and any('DUM' in t for t in entry.types_state_b):
            entry.params_state_b = [0.0, 0.0, '1']
        if entry.params_state_a and entry.subsection.header == 'dihedrals' and (not entry.params_state_b) \
                and entry.interaction_type in ('9', '4', '1'):
            if len(entry.params_state_a) > 3:
                assert len(entry.params_state_a) % 3 == 0
                leftover = entry.params_state_a[3:]
                entry.params_state_a = entry.params_state_a[:3]
                counter = 1
                while leftover:
                    new_entry = gml.EntryBonded(' '.join(str(x) for x in entry.content), self)
                    entry_location = entry.subsection.bkp_entries.index(entry)
                    entry.subsection.bkp_entries.insert(entry_location+counter, new_entry)
                    entry.subsection.bkp_entries[entry_location+counter].params_state_a = leftover[:3]
                    leftover = leftover[3:]
                    counter += 1
        if entry.params_state_a and entry.subsection.header == 'dihedrals' and entry.params_state_b \
                and entry.interaction_type in ('9', '4', '1'):
            if len(entry.params_state_a) > 3 or len(entry.params_state_b) > 3 \
                    or entry.params_state_a[2] != entry.params_state_b[2]:
                assert len(entry.params_state_a) % 3 == 0 and len(entry.params_state_b) % 3 == 0
                multiplicities_a = entry.params_state_a[2::3]
                multiplicities_b = entry.params_state_b[2::3]
                all_multiplicities = list(set(multiplicities_a + multiplicities_b))
                params_a = {entry.params_state_a[3*i+2]: entry.params_state_a[3*i:3*(i+1)]
                            for i in range(len(entry.params_state_a)//3)}
                params_b = {entry.params_state_b[3 * i + 2]: entry.params_state_b[3 * i:3 * (i + 1)]
                            for i in range(len(entry.params_state_b) // 3)}
                counter = 1
                m = all_multiplicities[-1]
                entry.params_state_a = params_a[m] if m in params_a.keys() else [0.0, 0.0, m]
                entry.params_state_b = params_b[m] if m in params_b.keys() else [0.0, 0.0, m]
                _ = all_multiplicities.pop()
                while all_multiplicities:
                    m = all_multiplicities[-1]
                    new_entry = gml.EntryBonded(' '.join(str(x) for x in entry.content), self)
                    entry_location = entry.subsection.bkp_entries.index(entry)
                    entry.subsection.bkp_entries.insert(entry_location + counter, new_entry)
                    entry.subsection.bkp_entries[entry_location + counter].params_state_a = params_a[m] \
                        if m in params_a.keys() else [0.0, 0.0, m]
                    entry.subsection.bkp_entries[entry_location + counter].params_state_b = params_b[m] \
                        if m in params_b.keys() else [0.0, 0.0, m]
                    counter += 1
                    _ = all_multiplicities.pop()
        if not entry.params_state_a and entry.subsection.header == 'dihedrals' and entry.params_state_b \
                and entry.interaction_type in ('9', '4', '1'):
            raise RuntimeError(f'Warning: in line {entry}, parameters were found for state A, but not for state B.'
                               f'Try to add parameters for types {entry.types_state_a}')
    
    def _check_parm_type(self):
        """
        Finds number codes for interaction type, e.g. CHARMM uses angletype '5' (urey-bradley)
        while Amber uses angletype '1' (simple harmonic); this will usually be one number
        but if subsections are merged, we can have mixed interaction types
        :return: list of strs, interaction types
        """
        types = []
        for entry in self.entries_bonded:
            if entry.interaction_type not in types:
                types.append(entry.interaction_type)
        return ['-1'] if not types else types

    def add_type_labels(self):
        for entry in self.entries_bonded:
            self._add_type_label(entry)

    @staticmethod
    def _add_type_label(entry):
        entry.read_types()
        entry.comment += " ; "
        entry.comment += " ".join(entry.types_state_a)
        if entry.types_state_b is not None:
            entry.comment += " ; "
            entry.comment += " ".join(entry.types_state_b)


class SubsectionParam(Subsection):
    """
    SubsectionParam contains force field parameters;
    should be included in SectionParam
    """
    n_atoms = {'pairtypes': 2, 'bondtypes': 2, 'constrainttypes': 2, 'angletypes': 3, 'dihedraltypes': 4,
               'nonbond_params': 2, 'defaults': 0, 'atomtypes': 1, 'implicit_genborn_params': 1, 'cmaptypes': 5}
    
    def __init__(self, content, section):
        super().__init__(content, section)
        self.atoms_per_entry = SubsectionParam.n_atoms[self.header]
        self.prmtypes = self._check_parm_type()
        self.label = '{}-{}'.format(self.header, self.prmtypes)
        self.ordering = {}
        if self.header == 'cmaptypes':
            self._process_cmap()
        
    def __repr__(self):
        if self.prmtypes[0] != '-1' or self.header not in ('atomtypes', 'implicit_genborn_params'):
            return "Subsection {} with interaction type {}".format(self.header, ' '.join(self.prmtypes))
        else:
            return "Subsection {}".format(self.header)
    
    def __add__(self, other):
        """
        Added for the purpose of merging subsections with
        identical headers
        :param other: other SubsectionParam instance
        :return: a new SubsectionParam instance resulting from the merger
        """
        if not isinstance(other, SubsectionParam):
            raise TypeError("{} is not a SubsectionParam instance".format(other))
        if self.header != other.header:
            raise TypeError("Cannot merge subsections with different headers: {} and {}".format(self.header,
                                                                                                other.header))
        return SubsectionParam(["[ {} ]\n".format(self.header)] + self.entries + other.entries, self.section)

    def get_entries_by_types(self, *types):
        found = []
        for entry in self.entries_param:
            if len(types) != len(entry.types):
                continue
            if all([query == tp for query, tp in zip(types, entry.types)]) \
                    or all([query == tp for query, tp in zip(types, entry.types[::-1])]):
                found.append(entry)
        if not found:
            raise RuntimeError(f"No entry found with types: {types}")
        return found

    def plot_term_energy(self, atomtypes, interaction_type=None, label=None, color='C0'):
        import matplotlib.pyplot as plt
        import numpy as np
        if isinstance(atomtypes, str):
            atomtypes = atomtypes.split()
        entries = self.get_entries_by_types(*atomtypes)
        assert all([e.subsection.header == entries[0].subsection.header for e in entries])
        if entries[0].subsection.header == "bondtypes":
            assert entries[0].interaction_type == '1' and len(entries) == 1
            x0, k = entries[0].params
            x = np.linspace(x0-0.025, x0+0.025, 100)
            plt.plot(x, 0.5 * k * (x-x0)**2, label=label, color=color)
        elif entries[0].subsection.header == "angletypes":
            assert len(entries) == 1
            if entries[0].interaction_type == '1':
                x0, k = entries[0].params
                x = np.linspace(x0-5, x0+5, 100)
                plt.plot(x, 0.5 * k * (x-x0)**2, label=label, color=color)
            elif entries[0].interaction_type == '5':
                raise NotImplementedError
        elif entries[0].subsection.header == "dihedraltypes":
            if entries[0].interaction_type == '9':
                x = np.linspace(-180, 180, 100)
                y = np.zeros(100)
                for entry in entries:
                    print(f'adding component {entry.params} from dihedral {entry.types} {entry.interaction_type}')
                    phi0, k, mult = entry.params
                    if mult == 0:
                        continue
                    y += k * (1 + np.cos(np.deg2rad(mult * x - phi0)))
                y -= np.min(y)
                plt.plot(x, y, label=label, color=color)
            else:
                raise NotImplementedError

    @property
    def entries_param(self):
        return [e for e in self.entries if isinstance(e, gml.EntryParam)]

    def sort(self):
        """
        In case we want to sort entries after some are added at the end of the section
        :return: None
        """
        for n, entry in enumerate(self.entries):
            if isinstance(entry, gml.EntryParam):
                key = tuple(str(entry).split()[:SubsectionParam.n_atoms[self.header]])
                if 'X' not in key:
                    self.ordering[key] = n
                else:
                    self.ordering[key] = n + 10**6
            else:
                self.ordering[str(entry)] = n - 10**6
        self.entries.sort(key=self._sorting_fn)

    def reorder_improper(self, types: tuple, new_order: str):
        try:
            entry = [ent for ent in self.get_entries_by_types(*types) if ent.interaction_type in '24'][0]
        except RuntimeError:
            return
        else:
            entry.types = (entry.types[int(new_order[0])], entry.types[int(new_order[1])],
                           entry.types[int(new_order[2])], entry.types[int(new_order[3])])

    def find_used_ff_params(self):
        used_parm_entries = []
        used_atomtypes_a = {a.type for mol in self.section.top.molecules for a in mol.atoms}
        used_atomtypes_b = {a.type_b for mol in self.section.top.molecules for a in mol.atoms if a.type_b is not None}
        used_atomtypes = used_atomtypes_a.union(used_atomtypes_b).union({'X'})
        for entry in self.entries_param:
            if all(tp in used_atomtypes for tp in entry.types):
                used_parm_entries.append(entry.identifier)
        return used_parm_entries

    def _combine_entries(self, other_subsection: "gml.SubsectionParam", overwrite: bool = False):
        """
        When merging parameter sets, makes sure there are no duplications, and in these cases resolves conflicts
        :param other_subsection: the gml.Subsection instance from which entries will be added to merge files
        :param overwrite: bool, whether to accept new parameters in case of a mismatch
        :return: None
        """
        # TODO check why cmap doesn't work?
        for entry in other_subsection.entries_param:
            existing = [e for e in self.entries_param if e.types == entry.types or e.types == entry.types[::-1]]
            if len(existing) == 1 or (len(existing) > 1 and self.header == 'dihedraltypes'):
                ex_entry = existing[0]
                if entry.params in [ee.params for ee in existing] and ex_entry.interaction_type == entry.interaction_type \
                        and entry.modifiers in [ee.modifiers for ee in existing]:
                    continue
                else:
                    if overwrite:
                        if len(existing) > 1:
                            # TODO fix this
                            raise RuntimeError(f'Found multiple entries matching the types of {entry.types} in section '
                                               f'{self.header}, cannot overwrite them all')
                        ex_entry.params = entry.params
                        ex_entry.interaction_type = entry.interaction_type
                        ex_entry.modifiers = entry.modifiers
                    else:
                        print(f"Found mismatch between original entry {str(ex_entry)} and new {entry}, keeping the "
                              f"original; if this is not desired, set overwrite=True")
            elif len(existing) == 0:
                if other_subsection.header != 'dihedraltypes':
                    self.add_entry(entry)
                else:
                    # a separate case if these are multiple dihedral entries to be added at once:
                    other_all = [e for e in other_subsection.entries_param if e.types == entry.types
                                 or e.types == entry.types[::-1]]
                    for dih_entry in other_all:
                        self.add_entry(dih_entry)
            else:
                raise RuntimeError(f"Found multiple entries matching the types of {entry.types} in section "
                                   f"{self.header}, make sure this is intentional")

    def _sorting_fn(self, entry: "gml.Entry") -> int:
        """
        Comments should go first, then we sort based on first, second,
        ... column of the section
        :param entry: Entry, entry to be sorted
        :return: int, ordering number
        """
        types = tuple(str(entry).split()[:SubsectionParam.n_atoms[self.header]]) if isinstance(entry, gml.EntryParam) \
            else str(entry)
        line_number = self.ordering[types]
        return line_number if not hasattr(entry, 'types_state_a') or 'X' not in entry.types \
            else line_number + len(self.ordering.keys())
    
    def _check_parm_type(self):
        """
        Finds number code for interaction type, e.g. CHARMM uses angletype '5' (urey-bradley)
        while Amber uses angletype '1' (simple harmonic)
        :return: str, interaction type
        """
        if self.header not in SubsectionParam.n_atoms.keys() or self.header in ['atomtypes', 'implicit_genborn_params']:
            return '-1'
        npar = SubsectionParam.n_atoms[self.header]
        types = list({entry.content[npar] if len(entry.content) > npar else entry.interaction_type
                      for entry in self.entries_param})
        return '-1' if not types else types

    def get_opt_dih(self, types=False):
        dopts = [entry for entry in self.entries_param if 'DIHOPT' in entry.comment]
        if not types:
            return [e.params[x] for e in dopts for x in [0, 1]]
        else:
            return [(*e.types, e.params[2]) for e in dopts]

    def get_opt_dih_indices(self):
        dopts = [entry for entry in self.entries_param if 'DIHOPT' in entry.comment]
        self.section.top.add_ff_params()
        indices = set()
        for i in dopts:  # TODO so far only works for one (first) molecule
            for j in self.section.top.molecules[0].get_subsection('dihedrals').entries_bonded:
                if i in j.params_state_a_entry:
                    indices.add(j.atom_numbers)
        return list(indices)

    def set_opt_dih(self, values):
        dopts = [entry for entry in self.entries_param if 'DIHOPT' in entry.comment]
        for e, ang, k in zip(dopts, values[::2], values[1::2]):
            e.params[0] = ang
            e.params[1] = k

    def _remove_symm(self, prefix):
        for n in range(len(self.entries)):
            if n >= len(self.entries):
                break
            # TODO fix this or rewrite from scratch
            if isinstance(self.entries[n], gml.EntryParam) and any([x.startswith(prefix)
                                                                    for x in self.entries[n].types]):
                limit = min(8, len(self.entries)-n-1)
                next_other = [q for q in range(1, limit) if isinstance(self.entries[n+q], gml.EntryParam)
                              and self.entries[n].types != self.entries[n+q]]
                if not next_other:
                    continue
                else:
                    next_other = next_other[0]
                for other in range(n+next_other, len(self.entries)):
                    if other < len(self.entries) and isinstance(self.entries[other], gml.EntryParam) \
                            and any([x.startswith(prefix) for x in self.entries[other].types]):
                        if self.entries[n].types == self.entries[other].types[::-1] \
                                and self.entries[n].params == self.entries[other].params:
                            _ = self.entries.pop(other)
    
    def _process_cmap(self):
        """
        Reads a multiline entry from the [ cmaptypes ] section
        and converts it into an array that can be later properly printed
        :return: None
        """
        new_entries = []
        current = []
        for e in self.entries:
            if isinstance(e, gml.EntryParam):
                if e.content[-1].endswith('\\'):
                    current.extend([x.rstrip('\\') for x in e.content])
                else:
                    current.extend([x.rstrip('\\') for x in e.content])
                    new_entry = ' '.join(current)
                    new_entries.append(gml.EntryParam(new_entry, self, processed=True))
                    current = []
            else:
                new_entries.append(e)
        self.entries = new_entries
        
        
class SubsectionAtom(Subsection):
    """
    SubsectionAtom contains definitions of all atoms in the molecule;
    should be contained in SectionMol
    """
    def __init__(self, content: list, section: "gml.SectionMol"):
        super().__init__(content, section)
        assert isinstance(self.section, gml.SectionMol)
        self.fstring = "{:>6}{:>11}{:>7}{:>7}{:>7}{:>7}{:>11}{:>11}   ; " + '\n'
        self.name_to_num, self.num_to_name, self.num_to_type, self.num_to_type_b = None, None, None, None

    @property
    def entries_atom(self):
        return [e for e in self.entries if isinstance(e, gml.EntryAtom)]

    def get_dicts(self, force_update=False):
        """
        dicts are not always needed and are costly to calculate,
        so only fill in the values when explicitly asked to
        :return: None
        """
        # TODO should they be @properties, called only when needed? EITHER WAY, REVISE
        if not self.name_to_num or force_update:
            self.name_to_num, self.num_to_name, self.num_to_type, self.num_to_type_b = self._mol_type_nums()

    def check_defined_types(self):
        typelist = self.section.top.defined_atomtypes
        for atom in self.entries_atom:
            if atom.type not in typelist:
                print(f'Couldn\'t find definition of atom type {atom.type} (atom {atom.num} in molecule '
                      f'{self.section.mol_name}) in parameters')
            if atom.type_b:
                if atom.type_b not in typelist:
                    print(f'Couldn\'t find definition of atom type {atom.type_b} (atom {atom.num} in molecule '
                          f'{self.section.mol_name}) in parameters')

    def _mol_type_nums(self):
        """
        Provides bindings between atomnumber and atomtype
        and vice versa for each molecule identified in
        the topology
        :return: tuple of dicts, each dict contains molname:(type:num) and
        molname:(num:type) bindings
        """
        name_to_num, num_to_name, num_to_type, num_to_type_b = {}, {}, {}, {}
        for entry in self.entries_atom:
            name_to_num[(entry.resid, entry.atomname)] = entry.num
            num_to_name[entry.num] = entry.atomname
            num_to_type[entry.num] = entry.type
            num_to_type_b[entry.num] = entry.type_b if entry.type_b is not None else entry.type
        return name_to_num, num_to_name, num_to_type, num_to_type_b

    
class SubsectionHeader(Subsection):
    """
    SubsectionHeader contains the [ moleculetype ] section;
    should be contained in SectionMol
    """
    def __init__(self, content, section):
        super().__init__(content, section)
        self.molname = [a.content[0] for a in self.entries if a.content][0]

    @property
    def entries_bonded(self):
        return []

    @property
    def entries_atom(self):
        return []


import os
from glob import glob
from shutil import copy2

import gromologist as gml
from typing import Optional, Iterable, Union


# TODO make top always optional between str/path and gml.Top


def generate_dftb3_aa(top: Union[str, "gml.Top"], selection: str, pdb: Optional[Union[str, "gml.Pdb"]] = None):
    """
    Prepares a DFT3B-compatible topology and structure, setting up amino acids
    for QM/MM calculations (as defined by the selection)
    :param top: gml.Top, a Topology object
    :param selection: str, a selection defining the residues to be modified
    :param pdb: gml.Pdb, a Pdb object (optional, alternatively can be an attribute of top)
    :return: None
    """
    top = gml.obj_or_str(top=top)
    pdb = gml.obj_or_str(pdb=pdb)
    special_atoms = {'N': -0.43, 'H': 0.35, 'HN': 0.35, 'C': 0.55, 'O': -0.47}
    atoms = top.get_atoms(selection)
    print("The following atoms were found:")
    for at in atoms:
        print(str(at))
    out = input("Proceed? (y/n)\n")
    if out.strip().lower() != 'y':
        return
    top.parameters.add_dummy_def('LA')
    mols = list(set(at.molname for at in atoms))
    for mol in mols:
        molecule = top.get_molecule(mol)
        current_atoms = [at for at in molecule.atoms if at in atoms]
        atom_indices = [at.num for at in current_atoms]
        current_bonds = molecule.get_subsection('bonds').entries_bonded
        for bond in current_bonds:
            if bond.atom_numbers[0] in atom_indices and bond.atom_numbers[1] in atom_indices:
                bond.interaction_type = '5'
                bond.params_state_a = []
        for atom in current_atoms:
            if atom.atomname not in special_atoms.keys():
                atom.charge = 0.0
            else:
                atom.charge = special_atoms[atom.atomname]
        cas = [at for at in current_atoms if at.atomname == 'CA']
        cbs = [at for at in current_atoms if at.atomname == 'CB']
        assert len(cas) == len(cbs)
        for ca, cb in zip(cas, cbs):
            molecule.add_vs2(ca.num, cb.num, 0.72, 'LIN', 'LA')
            molecule.add_constraint(ca.num, cb.num, 0.155)
        # TODO add vs2 to PDB for each chain that is affected
        cas_all, cbs_all = [at for at in atoms if at.atomname == 'CA'], [at for at in atoms if at.atomname == 'CB']
        if pdb is not None and top.pdb is None:
            top.add_pdb(pdb)

        for ca, cb in zip(cas_all, cbs_all):
            mol = top.get_molecule(ca.molname)
            for pdb_num_ca, last_atom in zip(mol._match_pdb_to_top(ca.num), mol._match_pdb_to_top(len(mol.atoms))):
                resid = top.pdb.atoms[pdb_num_ca].resnum
                chain = top.pdb.atoms[pdb_num_ca].chain
                top.pdb.add_vs2(resid, 'CA', 'CB', 'LIN', fraction=0.72, serial=last_atom, chain=chain)


def generate_gaussian_input(pdb: Union["gml.Pdb", str], directive_file: str, outfile: str = 'inp.gau', charge: int = 0,
                            multiplicity: int = 1, group_a: Optional[str] = None, group_b: Optional[str] = None):
    """
    From a .pdb file and an existing Gaussian input, produces a new .gau input
    with correct atom names, coordinates, and possibly fragment assignment
    :param pdb: gml.Pdb or str, the structure object/file containing the desired coordinates
    :param directive_file: str, an existing Gaussian input from which the %- and #-prefixed lines will be taken
    :param outfile: str, a file to which the new input will be written
    :param charge: int, charge of the system (by default 0)
    :param multiplicity: int, multiplicity of the system (by default 1)
    :param group_a: str, selection to define 1st fragment if the counterpoise correction is used
    :param group_b: str, selection to define 2nd fragment if the counterpoise correction is used
    :return: None
    """
    gau_content = [line for line in open(directive_file)]
    pdb = gml.obj_or_str(pdb=pdb)
    pdb.add_elements()
    with open(outfile, 'w') as outf:
        for line in [ln for ln in gau_content if ln.strip().startswith('%')]:
            outf.write(line)
        for line in [ln for ln in gau_content if ln.strip().startswith('#')]:
            outf.write(line)
        outf.write(f"\ngromologist input to gaussian\n\n{charge} {multiplicity}\n")
        if group_a is None and group_b is None:
            for atom in pdb.atoms:
                outf.write(f" {atom.element}   {atom.x}  {atom.y}  {atom.z}\n")
        elif group_a is not None and group_b is not None:
            for atom in pdb.get_atoms(group_a):
                outf.write(f" {atom.element}(Fragment=1)   {atom.x:8.3f}  {atom.y:8.3f}  {atom.z:8.3f}\n")
            for atom in pdb.get_atoms(group_b):
                outf.write(f" {atom.element}(Fragment=2)   {atom.x:8.3f}  {atom.y:8.3f}  {atom.z:8.3f}\n")
        else:
            raise RuntimeError('Specify either both group_a and group_b, or neither')
        outf.write("\n")


# TODO move REST2 preparation here

def parse_frcmod(filename: str) -> (dict, dict, dict, dict, dict, dict, dict):
    """
    Parses either an frcmod file with extra parameters, or a parm file with
    core FF parameters (both have slightly different formats for opening/closing sections)
    :param filename: str, name of the file to be read
    :return: tuple of dict, FF parameres in their respective formats
    """
    dat = True if filename.endswith('dat') else False
    content = open(filename).readlines()
    if any(['MOD4' in l and 'AC' in l for l in content]):
        raise RuntimeError("LJ type A/C not supported, terminating")
    content = content[1:] if dat else content
    atomtypes, bondtypes, angletypes, dihedraltypes, impropertypes, nonbonded, cmaptypes = {}, {}, {}, {}, {}, {}, {}
    cmapresol, cmapres, cmapvals, cmapread = None, [], [], False
    headers = ['MASS', 'BOND', 'ANGL', 'DIHE', 'IMPR', 'HBON', 'NONB', 'LJED', 'CMAP']
    identical_nonbonded = {}
    iterator = 0
    current = headers[iterator] if dat else None
    for nl, line in enumerate(content):
        if not dat:
            if any([line.strip().startswith(i) for i in headers]):
                current = line.strip()[:4]
                continue
            if current is None or not line.strip() or line.strip().startswith('#'):
                continue
        else:
            if not line.strip() and iterator < len(headers) - 3:
                iterator += 1
                current = headers[iterator]
                continue
            if line.strip() == "END":
                current = headers[-1]
        if current == 'BOND':
            if dat and '-' not in line[:5]:
                continue
            types = tuple(x.strip() for x in line[:5].split('-'))
            vals = tuple(float(x) for x in line[5:].split()[:2])
            bondtypes[types] = [vals[1] / 10, vals[0] * 200 * 4.184]
        elif current == 'ANGL':
            types = tuple(x.strip() for x in line[:8].split('-'))
            vals = tuple(float(x) for x in line[8:].split()[:2])
            angletypes[types] = [vals[1], vals[0] * 2 * 4.184]
        elif current == 'MASS':
            types = line.split()[0]
            mass = float(line.split()[1])
            if types in atomtypes.keys():
                atomtypes[types][0] = mass
            else:
                atomtypes[types] = [mass]
        elif current == 'NONB':
            if dat:
                try:
                   _ = float(line.split()[1])
                except:
                    if len(line.split()) >= 2 and all([t in atomtypes.keys() for t in line.split()]):
                        identical_nonbonded[line.split()[0]] = line.split()[1:]
                    continue
            else:
                if len(line.split()) < 3:
                    continue
            tps = line.split()[0]
            rmin = float(line.split()[1])
            eps = float(line.split()[2])
            types = [tps] if tps not in identical_nonbonded.keys() else [tps] + identical_nonbonded[tps]
            for atype in types:
                if atype in atomtypes.keys() and len(atomtypes[atype]) == 1:
                    atomtypes[atype].extend([rmin * 0.2 * 2 ** (-1 / 6), eps * 4.184])
                else:
                    atomtypes[atype] = [0, rmin * 0.2 * 2 ** (-1 / 6), eps * 4.184]
        elif current == 'LJED':
            if dat:
                if not(len(line.split()) > 1 and line.split()[0] in atomtypes.keys() and line.split()[1] in atomtypes.keys()):
                    continue
            types = tuple(line.split()[:2])
            vals = tuple(line.split()[2:])
            assert vals[0] == vals[2] and vals[1] == vals[3]
            nonbonded[types] = [float(vals[0]) * 0.2 * 2 ** (-1 / 6), float(vals[1]) * 4.184]
        elif current == 'DIHE':
            types = tuple(x.strip() for x in line[:12].split('-'))
            vals = tuple(float(x) for x in line[12:].split()[:4])
            entry = [vals[2], 4.184 * vals[1] / vals[0], int((vals[3] ** 2) ** 0.5)]
            if types in dihedraltypes.keys():
                dihedraltypes[types].extend(entry)
            else:
                dihedraltypes[types] = entry
        elif current == 'IMPR':
            types = tuple(x.strip() for x in line[:12].split('-'))
            vals = tuple(float(x) for x in line[12:].split()[:3])
            entry = [vals[1], 4.184 * vals[0], int((vals[2] ** 2) ** 0.5)]
            impropertypes[types] = entry
        elif current == 'CMAP':
            types = tuple('C N CT C N'.split()) # in case the format ever changes in the future
            if line.startswith('%FLAG'):
                if line.split()[1] == "CMAP_RESLIST":
                    cmapres = content[nl+1].split()
                elif line.split()[1] == "CMAP_RESOLUTION":
                    cmapresol = line.split()[2]
                elif line.split()[1] == "CMAP_PARAMETER":
                    cmapread = True
            elif cmapread:
                if not line.strip():
                    cmapread = False
                else:
                    cmapvals.extend(line.strip().split())
            if len(cmapvals) > 0 and len(cmapvals) == int(cmapresol) ** 2:
                cmapvals = [str(round(4.184 * float(i), 10)) for i in cmapvals]
                for res in cmapres:
                    cmaptypes[(types, res)] = [cmapresol, cmapvals]
                cmapresol, cmapres, cmapvals, cmapread = None, [], [], False

    # assert (all([len(val) == 3 for val in atomtypes.values()]))
    atomtypes = {key: val for key, val in atomtypes.items() if len(val) == 3}
    non_atomtypes = [key for key, val in atomtypes.items() if len(val) != 3]
    if non_atomtypes:
        print(f"skipping atomtypes {non_atomtypes}, missing LJ parameters")
    return atomtypes, bondtypes, angletypes, dihedraltypes, impropertypes, nonbonded, cmaptypes


def load_frcmod(top: Union[str, "gml.Top"], filename: str, return_cmap: bool = False, pro_atoms: Optional[dict] = None)\
        -> Optional[dict]:
    """
    Loads an .frcmod file into an existing topology. Can be also launched as
    gml.Top().load_frcmod(...)
    :param top: str or gml.Top, existing gmx topology
    :param filename: str, name of the frcmod file to load
    :param return_cmap: bool, if set to True will return cmaptypes
    :return: None or dict, depending on return_cmap
    """
    top = gml.obj_or_str(top)
    atomtypes, bondtypes, angletypes, dihedraltypes, impropertypes, nonbonded, cmaptypes = parse_frcmod(filename)
    for k in cmaptypes.keys():
        try:
            cc = [adata[1] for adata in pro_atoms[k[1]] if adata[0] == "C"][0]
            na = [adata[1] for adata in pro_atoms[k[1]] if adata[0] == "N"][0]
            ca = [adata[1] for adata in pro_atoms[k[1]] if adata[0] == "CA"][0]
        except KeyError:
            cc, na, ca = "C", "N", "CA"
        cmaptypes[k].append(f'{cc} {na} {ca} {cc} {na}'.split())
    params = top.parameters
    for at in atomtypes.keys():
        params.add_atomtype(at, *atomtypes[at], action_default='r')
    for b in bondtypes.keys():
        params.add_bonded_param(b, bondtypes[b], 1, action_default='r')
    for a in angletypes.keys():
        params.add_bonded_param(a, angletypes[a], 1, action_default='r')
    for d in dihedraltypes.keys():
        # TODO add wildcards at the end?
        params.add_bonded_param(d, dihedraltypes[d], 9, action_default='r')
    for i in impropertypes.keys():
        params.add_bonded_param(i, impropertypes[i], 4, action_default='r')
    for n in nonbonded.keys():
        try:
            params.add_nbfix(*n, new_sigma=nonbonded[n][0], new_epsilon=nonbonded[n][1])
        except KeyError:
            print(f"Skipping NBFIX {n} as at least one of the types is not defined; if you want to keep it, "
                  "create/load the type and run this command again.")
    for c in cmaptypes.keys():
        params.add_bonded_param(cmaptypes[c][-1], [c[1], cmaptypes[c][:-1]], 1, action_default='a')
    if return_cmap:
        return cmaptypes
    else:
        return None


def read_lib(lib: str) -> (dict, dict, dict):
    """
    Reads a .lib file with residue definitions
    :param lib: str, name of the .lib file
    :return: tuple of dict, dictionary with atoms, bonds, and inter-residue connectors
    """
    curr_resname = None
    atoms = {}
    bonds = {}
    connector = {}
    reading_atoms = False
    reading_bonds = False
    content = [line for line in open(lib) if line.strip()]
    for n, ln in enumerate(content):
        if not ln.startswith('!'):
            if reading_atoms:
                atoms[curr_resname].append((ln.strip().split()[0].strip('"'), ln.strip().split()[1].strip('"'),
                                            float(ln.strip().split()[7]), int(ln.strip().split()[5])))
            elif reading_bonds:
                bonds[curr_resname].append((int(ln.strip().split()[0]), int(ln.strip().split()[1])))
        if ln.startswith('!'):
            if len(ln.strip('!').split()[0].split('.')) < 3:
                continue
            else:
                reading_bonds, reading_atoms = False, False
                if ln.strip('!').split()[0].split('.')[3] == 'atoms':
                    reading_atoms = True
                    curr_resname = ln.strip('!').split()[0].split('.')[1]
                    atoms[curr_resname] = []
                    bonds[curr_resname] = []
                    connector[curr_resname] = []
                elif ln.strip('!').split()[0].split('.')[3] == 'connectivity':
                    reading_bonds = True
                elif ln.strip('!').split()[0].split('.')[3] == 'connect':
                    connector[curr_resname].append(int(content[n + 1].strip()))
    atoms = {k: v for k, v in atoms.items() if 'BOX' not in k}
    bonds = {k: v for k, v in bonds.items() if 'BOX' not in k}
    connector = {k: v for k, v in connector.items() if 'BOX' not in k}
    return atoms, bonds, connector


def write_rtp(atoms: dict, bonds: dict, connector: dict, outfile: str = "new.rtp", ff='amber',
              impropers: Optional[dict] = None, cmap: Optional[dict] = None):
    """
    Writes an .rtp file given all per-residue dictionary with topology, extra dihedrals/impropers, CMAP etc.
    :param atoms: dict of tuple, atom names/types and their ID/charge
    :param bonds: dict of tuples, 1-based indices of pairs defining bonds
    :param connector: dict of tuples, connecting atoms from -1 or +1 residue in polymers
    :param outfile: str, to which file output should be written
    :param ff: str, 'amber' or 'charmm' to set correct [ bondedtypes ] (default interaction types)
    :param impropers: dict of tuples, atom names that should be involved in improper dihedrals
    :param cmap: dict of tuples, atom names that should be involved in the cmap correction
    :return: None
    """
    if ff.lower() not in ['amber', 'charmm']:
        raise RuntimeError("Only Amber and CHARMM are currently supported")
    btypes = '11941310' if ff == 'amber' else '15921310'
    print(f"Setting [ bondedtypes ] in {outfile} file for the {ff}-type force field, please make sure this is right")
    with open(outfile, 'w') as out:
        out.write(f"[ bondedtypes ]\n{' '.join(btypes)}\n\n")
        for res in atoms.keys():
            out.write(f"[ {res} ]\n [ atoms ]\n")
            for at in atoms[res]:
                out.write(f"  {at[0]:4s}   {at[1]:4s}          {at[2]:8.5f}     {at[3]:3d}\n")
            out.write(f" [ bonds ]\n")
            for bd in bonds[res]:
                out.write(f"  {atoms[res][bd[0] - 1][0]:4s}   {atoms[res][bd[1] - 1][0]:4s}\n")
            if len(connector[res]) > 0 and connector[res][0] > 0:
                atomlist = [at[0] for at in atoms[res]]
                is_prot = True if 'CA' in atomlist else False
                is_na = True if "O4'" in atomlist else False
                if is_prot:
                    out.write(f"  -C  {atoms[res][connector[res][0] - 1][0]}\n")
                elif is_na:
                    out.write(f"  -O3'  {atoms[res][connector[res][0] - 1][0]}\n")
            if impropers is not None:
                if res in impropers.keys():
                    out.write(f" [ impropers ]\n")
                    for imp in impropers[res]:
                        out.write(f" {imp[0]:5s} {imp[1]:5s} {imp[2]:5s} {imp[3]:5s}\n")
            if cmap is not None:
                if res in cmap.keys():
                    out.write(f" [ cmap ]\n")
                    for cmp in cmap[res]:
                        out.write(f" {cmp[0]:5s} {cmp[1]:5s} {cmp[2]:5s} {cmp[3]:5s} {cmp[4]:5s}\n")
            out.write("\n\n")


def dict_filter(indict: dict, restype: str) -> dict:
    """
    Filters dictionaries based on whether they belong to DNA, RNA or protein
    :param indict: dict, dictionary with keys that are residue names
    :param restype: str, type of residues (DNA, RNA, or anything else for protein)
    :return: dict, the filtered dictionary
    """
    nucres = gml.Pdb.nucl_map.keys()
    if restype == "DNA":
        return {k: v for k, v in indict.items() if k in nucres and 'D' in nucres}
    elif restype == "RNA":
        return {k: v for k, v in indict.items() if k in nucres and 'D' not in nucres}
    else:
        return {k: v for k, v in indict.items() if k not in nucres}


def fix_rtp(rtp_dict: dict, impr: bool = False, rna: bool = False) -> dict:
    """
    Makes Gromacs-specific changes in .rtp data, e.g. adjusts atom names, terminal atom naming,
    hydrogen numbering, RNA residue names, improper type order etc.
    :param rtp_dict: dict, dictionary with .rtp data (atoms or impropers)
    :param impr: bool, whether the entry contains improper dihedral data
    :param rna: bool, whether the entry represents RNA residues
    :return: dict, the dictionary with necessary modifications
    """
    to_copy = {}
    for res in rtp_dict.keys():
        if not rna:
            if "ILE" in res:
                for n, ent in enumerate(rtp_dict[res]):
                    if ent[0] == "CD1":
                        rtp_dict[res][n] = ('CD', *ent[1:])
                    elif ent[0].startswith("HD1"):
                        rtp_dict[res][n] = (ent[0].replace("HD1", "HD"), *ent[1:])
            if res.startswith("C") and res[1:] in gml.Pdb.prot_map.keys():
                if not impr:
                    for n, ent in enumerate(rtp_dict[res]):
                        if ent[0] == "O":
                            rtp_dict[res][n] = ('OC1', *ent[1:])
                        elif ent[0] == "OXT":
                            rtp_dict[res][n] = ('OC2', *ent[1:])
                else:
                    for n, ent in enumerate(rtp_dict[res]):
                        for m, atname in enumerate(ent):
                            if atname == "O":
                                rtp_dict[res][n][m] = 'OC2'
                            elif atname == "OXT":
                                rtp_dict[res][n][m] = 'OC1'
            for mid in ["A", "B", "G", "D", "E", "G1"]:
                if f"H{mid}3" in [e[0] for e in rtp_dict[res]] and f"H{mid}1" not in [e[0] for e in rtp_dict[res]]:
                    for n, ent in enumerate(rtp_dict[res]):
                        if ent[0] == f"H{mid}3":
                            rtp_dict[res][n] = (f'H{mid}1', *ent[1:])
            if impr:
                for n, ent in enumerate(rtp_dict[res]):
                    for m, atname in enumerate(ent):
                        if rtp_dict[res][n][m] == '-M':
                            rtp_dict[res][n][m] = '-C'
                        elif rtp_dict[res][n][m] == '+M':
                            rtp_dict[res][n][m] = '+N'
        if rna:
            for res in rtp_dict.keys():
                if res[0] in 'CAGU':
                    to_copy[res] = 'R' + res
                if impr:
                    continue
                for n, ent in enumerate(rtp_dict[res]):
                    try:
                        _ = ent[0]
                    except TypeError:
                        continue
                    if ent[0] == "OP1":
                        rtp_dict[res][n] = ('O1P', *ent[1:])
                    elif ent[0] == "OP2":
                        rtp_dict[res][n] = ('O2P', *ent[1:])
                    elif ent[0] == "H5''":
                        rtp_dict[res][n] = ("H5'2", *ent[1:])
                    elif ent[0] == "H5'":
                        rtp_dict[res][n] = ("H5'1", *ent[1:])
                    elif ent[0] == "H2'":
                        rtp_dict[res][n] = ("H2'1", *ent[1:])
                    elif ent[0] == "HO2'":
                        rtp_dict[res][n] = ("HO'2", *ent[1:])
                    elif ent[0] == "HO3'":
                        rtp_dict[res][n] = ("H3T", *ent[1:])
                    elif ent[0] == "HO5'":
                        rtp_dict[res][n] = ("H5T", *ent[1:])
    for cop in to_copy.keys():
        rtp_dict[to_copy[cop]] = rtp_dict[cop]
    for clean in to_copy.keys():
        del rtp_dict[clean]
    return rtp_dict


def find_leap_files(ambdir: str, filetype: str, leaprc: str, extras: Optional[list] = None):
    content = [line.strip() for line in open(leaprc)]
    extras = extras if extras is not None else []
    pref = os.path.sep if leaprc.startswith(os.path.sep) else ''
    reldir = pref + os.path.sep.join(leaprc.split(os.path.sep)[:-1]) + pref
    if filetype == 'lib':
        files = [line.split()[-1] for line in content if
                 len(line.split()) >= 2 and "loadoff" in [w.lower() for w in line.split()]] + extras
    elif filetype == 'parm':
        files = [line.split()[-1] for line in content if
                 len(line.split()) >= 2 and "loadamberparams" in [w.lower() for w in line.split()]] + extras
    else:
        raise RuntimeError('select "parm" or "lib"')
    files_in_ambdir = [ambdir + f'/{filetype}/' + f for f in files]
    files_in_reldir = [reldir + '/' + f for f in files]
    print(files_in_reldir)
    found_files = [f for f in files_in_ambdir + files_in_reldir if os.path.exists(f)]
    return found_files


def amber2gmxFF(leaprc: str, outdir: str, amber_dir: Optional[str] = None, base_ff: Optional[str] = None,
                extra_files: Optional[list] = None):
    """
    Reads a .leaprc file and all parameter dependencies from Amber to convert into a Gromacs .ff dir
    Files that should be copied manually: watermodels.dat and tip*itp, .hdb, .tdb and .arn
    :param leaprc: str, a file that sources dependencies from which the .ff will be created
    :param outdir: str, a new .ff directory that will contain the Gromacs-compatible files
    :param amber_dir: str, Abs path to the dir containing Amber prep, parm, lib directories if `leaprc` is a local file
    :param base_ff: str, name of the source .ff directory that will be used for auxilliary files
    :param extra_files: list of str, additional files that should be included (e.g. frcmod.tip3p)
    :return: None
    """
    orig_dir = os.path.sep.join(leaprc.split(os.path.sep)[:-1]) + os.path.sep if os.path.sep in leaprc else ''
    if amber_dir is not None:
        amb = amber_dir
    else:
        amb = f'{orig_dir}../'
    extra_files = extra_files if extra_files is not None else []
    extra_libs = [pth for pth in extra_files if pth.endswith('lib')]
    extra_dats = [pth for pth in extra_files if pth.endswith('dat') or 'frcmod' in pth.split(os.path.sep)[-1]]
    extra_prep = [pth for pth in extra_files if pth.endswith('in') or pth.endswith('prep')]
    libs = find_leap_files(amb, 'lib', leaprc, extra_libs)
    dats = find_leap_files(amb, 'parm', leaprc, extra_dats)
    pro_atoms, pro_bonds, pro_connectors = {}, {}, {}
    dna_atoms, dna_bonds, dna_connectors = {}, {}, {}
    rna_atoms, rna_bonds, rna_connectors = {}, {}, {}
    impropers = {}
    for prep in glob(amb + "/prep/all*.in") + glob(amb + "/prep/nuc*.in") + extra_prep:
        impropers.update(read_prep_impropers(prep))
    print('\n')
    for lib in libs:
        print(f"***** Adding residues from {lib} *****")
        a, b, c = gml.read_lib(lib)
        pro_atoms.update(fix_rtp(dict_filter(a, 'protein')))
        pro_bonds.update(dict_filter(b, 'protein'))
        pro_connectors.update(dict_filter(c, 'protein'))
        dna_atoms.update(dict_filter(a, 'DNA'))
        dna_bonds.update(dict_filter(b, 'DNA'))
        dna_connectors.update(dict_filter(c, 'DNA'))
        rna_atoms.update(fix_rtp(dict_filter(a, 'RNA'), rna=True))
        rna_bonds.update(fix_rtp(dict_filter(b, 'RNA'), rna=True))
        rna_connectors.update(fix_rtp(dict_filter(c, 'RNA'), rna=True))
    print('\n')
    pro_impropers = fix_rtp(dict_filter(impropers, 'protein'), impr=True)
    dna_impropers = dict_filter(impropers, 'DNA')
    rna_impropers = fix_rtp(dict_filter(impropers, 'RNA'), impr=True, rna=True)
    new_top = gml.Top(amber=True)
    cmaptypes, rtp_cmap = {}, {}
    print('\n')
    for dat in dats:
        print(f"***** Adding parameters from {dat} *****")
        cmaptypes.update(load_frcmod(new_top, dat, return_cmap=True, pro_atoms=pro_atoms))
    print('\n')
    for k in cmaptypes.keys():
        rtp_cmap[k[1]] = [f'-C N CA C +N'.split()]
    new_top = reorder_amber_impropers(new_top)
    outdir = outdir + '.ff' if not outdir.endswith('.ff') else outdir
    atomtypes = read_addAtomTypes(leaprc)
    os.mkdir(outdir)
    os.chdir(outdir)
    for atype in new_top.parameters.atomtypes.entries_param:
        if atype.modifiers[0] == '0' and atype.types[0] in atomtypes.keys():
            atype.modifiers[0] = str(atomtypes[atype.types[0]][0])
        if atype.modifiers[1] == '0' and atype.types[0] in atomtypes.keys():
            atype.modifiers[1] = str(atomtypes[atype.types[0]][1])
    new_top.save_top('forcefield.itp', split=True)
    gml.write_rtp(pro_atoms, pro_bonds, pro_connectors, 'aminoacids.rtp', impropers=pro_impropers, cmap=rtp_cmap)
    if dna_atoms:
        gml.write_rtp(dna_atoms, dna_bonds, dna_connectors, 'dna.rtp', impropers=dna_impropers)
    if rna_atoms:
        gml.write_rtp(rna_atoms, rna_bonds, rna_connectors, 'rna.rtp', impropers=rna_impropers)
    gmx_dir = new_top.gromacs_dir
    if not gmx_dir:
        print("Gromacs directory not found. Please move the additional files (.hdb, solvent .itp etc.) manually")
    else:
        if base_ff is None:
            base_ff = gmx_dir + '/amber99.ff'
        else:
            if not base_ff.startswith('/'):
                base_ff = gmx_dir + f'/{base_ff}'
        wildcards_to_copy = ['tip*', 'spc*', '*r2b', '*tdb', '*arn', '*hdb']
        files_to_copy = []
        for wildcard in wildcards_to_copy:
            files_to_copy.extend(glob(f'{base_ff}/{wildcard}'))
        print(f"Copying accessory files (.hdb, .atp, solvent .itp etc.) from {base_ff}\n")
        for file_to_copy in files_to_copy:
            copy2(file_to_copy, '.')
    with open('atomtypes.atp', 'w') as outfile:
        for atype in new_top.parameters.atomtypes.entries_param:
            outfile.write(f"{atype.types[0]:12s} {float(atype.modifiers[1]):12.5f}\n")
    new_top.find_undefined_types()
    # TODO set up watermodels.dat


def reorder_amber_impropers(new_top: "gml.Top") -> "gml.Top":
    """
    Modifying improper dihedral order, empirically checked against GMX FFs
    when errors come up
    :param new_top: gml.Top, a topology to process
    :return: gml.Top
    """
    try:
        _ = new_top.parameters.dihedraltypes
    except KeyError:
        pass
    else:
        new_top.parameters.dihedraltypes.reorder_improper(('CB', 'CT', 'C*', 'CW'), '1203')
        new_top.parameters.dihedraltypes.reorder_improper(('CT', 'CW', 'CC', 'NB'), '0213')
        new_top.parameters.dihedraltypes.reorder_improper(('CB', 'N2', 'CA', 'NC'), '0321')
        new_top.parameters.dihedraltypes.reorder_improper(('CB', 'C5', 'N*', 'CT'), '3201')
        new_top.parameters.dihedraltypes.reorder_improper(('CB', 'CP', 'N*', 'CT'), '3201')
        new_top.parameters.dihedraltypes.reorder_improper(('C', 'C4', 'N*', 'CT'), '3201')
        new_top.parameters.dihedraltypes.reorder_improper(('C', 'CS', 'N*', 'CT'), '3201')
        new_top.parameters.dihedraltypes.reorder_improper(('C4', 'N2', 'CA', 'NC'), '1203')
        new_top.parameters.dihedraltypes.reorder_improper(('N2', 'NA', 'CA', 'NC'), '1320')
    return new_top


def read_addAtomTypes(textfile: str) -> dict:
    text = open(textfile).readlines()
    reading, brack = False, 0
    element_properties = {
        "H": (1, 1.008), "O": (8, 15.999), "C": (6, 12.011), "N": (7, 14.007), "S": (16, 32.06), "P": (15, 30.974),
        "F": (9, 18.998), "Cl": (17, 35.45), "Br": (35, 79.904), "I": (53, 126.90), "Mg": (12, 24.305),
        "Ca": (20, 40.078)
    }
    types = {}
    for line in text:
        if line.strip().startswith('addAtomTypes'):
            reading = True
        if reading:
            brack += line.count('{')
            brack -= line.count('}')
        else:
            continue
        if brack == 0:
            reading = False
        data = line.strip().strip('{}').strip().split()
        if len(data) == 3:
            types[data[0].strip('"')] = data[1].strip('"')
    for k, v in types.items():
        if v and v in element_properties.keys():
            types[k] = element_properties[v]
        else:
            types[k] = (0, 0.0)
    return types


def read_prep_impropers(prepfile: str) -> dict:
    """
    Reads improper dihedrals from a specified file in the leap/prep directory
    :param prepfile: str, input file for reading
    :return: dict of lists, each dict entry corresponds to a residue, the list contains 4-lists of atomtypes
    """
    impropers = {}
    current = None
    reading = False
    content = [line for line in open(prepfile)]
    for line in content:
        if len(line.split()) > 2:
            if line.strip().split()[1] == "INT":
                current = line.strip().split()[0]
        if line.strip() == "IMPROPER":
            reading = True
        if line.strip() == "DONE":
            reading = False
        if reading and current is not None and line.strip():
            if current not in impropers.keys():
                impropers[current] = []
            if len(line.split()) == 4:
                types = line.strip().split()
                impropers[current].append(types)
    return impropers


def calc_Coulomb_force(top: Union[str, "gml.Top"], pdb: Union[str, "gml.Pdb"], force_on: str, force_from: str):
    """
    Calculates Coulomb forces exerted by one subset of atoms on each atom in another subset. Does NOT take into accout
    scaled 1-4 interactions.
    :param top: gml.Top or str, topology
    :param pdb: gml.Pdb or str, structure
    :param force_on: selection for which the forces will be calculated
    :param force_from: selection of atoms that are exerting the forces
    :return: Nx3 np.array, where N is the length of the force_on selection
    """
    import numpy as np
    k_coul = 138.9118  # in kJ/mol nm e²
    pdb = gml.obj_or_str(pdb=pdb)
    top = gml.obj_or_str(top=top)
    fon_indices = pdb.get_atom_indices(force_on)
    ffr_indices = pdb.get_atom_indices(force_from)
    topat = top.atoms
    fon_charges = [a.charge for a in [topat[i] for i in fon_indices]]
    ffr_charges = np.array([a.charge for a in [topat[i] for i in ffr_indices]])
    fon_coords = np.array([pdb.get_coords(force_on)])[0]/10
    ffr_coords = np.array([pdb.get_coords(force_from)])[0]/10
    vec_matrix = np.zeros((len(fon_indices), len(ffr_indices), 3))
    for i in range(len(fon_indices)):
        for j in range(len(ffr_indices)):
            vec_matrix[i, j, :] = fon_coords[i] - ffr_coords[j]
    dist_matrix = np.linalg.norm(vec_matrix, axis=2)[..., None]
    geom_matrix = vec_matrix / dist_matrix**3
    force_matrix = np.zeros((len(fon_indices), 3))
    for i in range(len(fon_indices)):
        force_matrix[i, :] = k_coul * fon_charges[i] * np.sum(ffr_charges[..., None] * geom_matrix[i, :, :], axis=0)
    return force_matrix


def calc_LJ_force(top: Union[str, "gml.Top"], pdb: Union[str, "gml.Pdb"], force_on: str, force_from: str):
    """
    Calculates Coulomb forces exerted by one subset of atoms on each atom in another subset. Does NOT take into accout
    scaled 1-4 interactions.
    :param top: gml.Top or str, topology
    :param pdb: gml.Pdb or str, structure
    :param force_on: selection for which the forces will be calculated
    :param force_from: selection of atoms that are exerting the forces
    :return: Nx3 np.array, where N is the length of the force_on selection
    """
    import numpy as np
    pdb = gml.obj_or_str(pdb=pdb)
    top = gml.obj_or_str(top=top)
    fon_indices = pdb.get_atom_indices(force_on)
    ffr_indices = pdb.get_atom_indices(force_from)
    topat = top.atoms
    sigma_matrix = np.array([[top.parameters.sigma_ij(topat[i], topat[j]) for i in ffr_indices] for j in fon_indices])
    epsilon_matrix = np.array(
        [[top.parameters.epsilon_ij(topat[i], topat[j]) for i in ffr_indices] for j in fon_indices])
    fon_coords = np.array([pdb.get_coords(force_on)])[0]/10
    ffr_coords = np.array([pdb.get_coords(force_from)])[0]/10
    vec_matrix = np.zeros((len(fon_indices), len(ffr_indices), 3))
    for i in range(len(fon_indices)):
        for j in range(len(ffr_indices)):
            vec_matrix[i, j, :] = fon_coords[i] - ffr_coords[j]
    dist_matrix = np.linalg.norm(vec_matrix, axis=2)[..., None]
    geom_matrix = vec_matrix / dist_matrix ** 2
    s6_matrix = (sigma_matrix / dist_matrix[:, :, 0]) ** 6
    force_matrix = np.zeros((len(fon_indices), 3))
    for i in range(len(fon_indices)):
        force_matrix[i, :] = 24 * np.sum(epsilon_matrix[i, :][..., None] * geom_matrix[i, :, :] * (2 * s6_matrix**2 - s6_matrix)[i, :][..., None], axis=0)
    return force_matrix



import os

import numpy as np

import gromologist as gml
from typing import Union, Optional, TypeVar, Tuple
from itertools import combinations_with_replacement
from concurrent.futures import ProcessPoolExecutor
from copy import deepcopy

gmlMod = TypeVar("gmlMod", bound="Mod")


class Mod:

    def __init__(self, td: "gml.ThermoDiff", top: Union[str, gml.Top], structure: str, selections: Optional[list] = None):
        """
        Base class to work with parameter modifications.
        The class is actually used for the calculation of original
        (unmodified) quantities, and is subclassed by ModParam,
        ModNbfix and ModAtom.
        :param top: list of Top objects corresponding to the modified topology
        :param structure: str, path to a structure file (.gro or .pdb) compatible with the topology
        :param selections: list of str, selections that will define atoms to be modified
        """
        td.counter += 1
        self.counter = td.counter
        self.td = td
        # TODO rethink if topologies shouldn't be stored in text files instead of in memory (makes it hard to run massive analyses locally)
        if isinstance(top, str):
            self.top = gml.Top(top)
        elif isinstance(top, gml.Top):
            self.top = top
        else:
            raise RuntimeError('"top" has to be either a path to file or a gml.Top object')
        self.data = None
        self.diff = None
        self.structure = structure
        self.topname = ''
        self.trajs = []
        self.dpar = 0.001
        self.sigma, self.chg, self.eps, self.nbs, self.nbe, self.dih, self.ang = [''] * 7
        self.atoms = [self.top.get_atoms(sel) for sel in selections]

    def __str__(self) -> str:
        """
        string representation of a Mod object, should be unique
        :return: mod one-letter code + Mod index, 1-based
        """
        return "{}-{}".format(''.join([self.sigma, self.chg, self.eps, self.nbs, self.nbe, self.dih, self.ang]),
                              self.counter)

    def __eq__(self, other: gmlMod) -> bool:
        """
        assumes unique numbering of instances
        :param other: another Mod (or subclass) instance to compare with
        :return: bool
        """
        return str(self) == str(other)


    def save_mod(self, path: str, name: str):
        """
        wraps Top's save_mod to catch the file name
        :param path: str, filename to save .top to
        :param name: str, name of the file
        :return: None
        """
        self.topname = path
        final_name = self.topname + '/' + name + '-' + self.top.fname.split('/')[-1]
        if not os.path.exists(final_name):
            self.top.save_top(final_name)
        else:
            print(f'file {final_name} already exists, leaving as it is')

    def goto_mydir(self):
        """
        shortcut to the ModAtom's working directory
        :return: None
        """
        os.chdir(self.topname)


class ModAtom(Mod):
    def __init__(self, td: "gml.ThermoDiff", top: Union[str, gml.Top], structure: str, selections: list, changes: str):
        """
        Subclass to work with modified atomic params (that is,
        sigma and epsilon defined for individual types, or charge defined
        for individual atoms).
        The class takes in a set of atoms that all correspond
        to a single atom type (has to be the same for all atoms);
        these atoms' type will be cloned from "type" to "Ytype",
        and the corresponding sigma/epsilon/charge parameters will be modified.
        :param top: a Top object corresponding to the modified topology
        :param selections: list of str, selections that will define atoms to be modified
        :param changes: single character denoting the change to be performed
        """
        super(ModAtom, self).__init__(td, top, structure, selections)
        self.sigma = 's' if 's' in changes else ''
        self.eps = 'e' if 'e' in changes else ''
        self.chg = 'c' if 'c' in changes else ''
        assert len(self.atoms) == 1
        self.atoms = self.atoms[0]
        if not (self.chg or self.eps or self.sigma):
            raise ValueError("no mode selected appropriate for an ModAtom object")
        self.types_are_consistent()
        self.type = self.atoms[0].type
        if not self.chg:
            self.top.parameters.clone_type(atomtype=self.type, prefix='Y')
        self.mod_sigma_eps()
        self.mod_chg()

    def __repr__(self):
        """
        string representation of a Mod object, should be unique
        :return: mod one-letter code + Mod index, 1-based
        """
        return "{}-{}, types {}".format(
            ''.join([self.sigma, self.chg, self.eps, self.nbs, self.nbe, self.dih, self.ang]),
            self.counter, self.type)

    def mod_chg(self):
        """
        looks for a line in topology section 'atoms'
        and modifies it as needed
        :return: None
        """
        for atom in self.atoms:
            if "TD-MODDED" not in atom.comment:
                atom.comment = atom.comment + " ; TD-MODDED"
                atom.mass_b = atom.mass
                if self.chg:
                    atom.type_b = atom.type
                    atom.charge_b = atom.charge + self.dpar
                if self.sigma or self.eps:
                    atom.type_b = 'Y' + atom.type
                    atom.charge_b = atom.charge

    def mod_sigma_eps(self):
        """
        looks for a line in topology section 'atomtypes'
        and modifies it as needed
        :return: None
        """
        subsection = self.top.parameters.get_subsection('atomtypes')
        for e in subsection.entries_param:
            if e.types[0] == 'Y' + self.type:
                if self.sigma:
                    e.params[0] += self.dpar
                if self.eps:
                    e.params[1] += self.dpar

    def types_are_consistent(self):
        """
        if many atoms are passed in a single input line, need to make sure
        they are all of the same type
        :return: True if all are of the same type, False otherwise
        """
        consistent = all([a.type == self.atoms[0].type for a in self.atoms])
        if not consistent:
            raise ValueError(
                "atoms within a single charge, sigma or epsilon modification need to have consistent types. "
                "To change atoms with different types, define multiple modifications")


class ModNbfix(Mod):
    def __init__(self, td: "gml.ThermoDiff", top: Union[str, gml.Top], structure: str, selections: list, changes: str):
        """
        Subclass to work with modified NBFIX params (pairwise corrections,
        sigma and epsilon defined for pairs of types).
        The class allows two lists (actually, tuples) of atoms that correspond
        to two atom types (atomtypes have to be consistent within each tuple);
        these atoms' types will be cloned from (type1, type2) to (Xtype1,
        Ytype2), and the corresponding NBFIX parameters will be modified.
        The modification can be applied to all atoms bearing the specific type
        as well as only a subgroup of these.
        :param top: a Top object corresponding to the modified topology
        :param selections: list of str, selections that will define atoms to be modified
        :param changes: single character denoting the change to be performed
        """
        super(ModNbfix, self).__init__(td, top, structure, selections)
        self.nbs = 'n' if 'n' in changes else ''
        self.nbe = 'm' if 'm' in changes else ''
        if not (self.nbs or self.nbe):
            raise ValueError("no mode selected appropriate for an ModParam object")
        self.types = (self.atoms[0][0].type, self.atoms[1][0].type)
        if self.types[0] == self.types[1]:
            self.prefixes = ['Y', 'Y']
        else:
            self.prefixes = ['Q', 'Y']
        self.types_are_consistent()
        for type, prefix in zip(self.types, self.prefixes):
            if prefix + type not in self.top.defined_atomtypes:
                self.top.parameters.clone_type(atomtype=type, prefix=prefix)
        # we mod the first type
        self.prefix_type(self.atoms[0], self.types[0], self.prefixes[0])
        # and then second, if different than first
        if self.types[0] != self.types[1]:
            self.prefix_type(self.atoms[1], self.types[1], self.prefixes[1])
        self.mod_nb_sigma_eps()

    def __repr__(self):
        """
        string representation of a Mod object, should be unique
        :return: mod one-letter code + Mod index, 1-based
        """
        return "{}-{}, types {}".format(
            ''.join([self.sigma, self.chg, self.eps, self.nbs, self.nbe, self.dih, self.ang]),
            self.counter, self.types)

    def mod_nb_sigma_eps(self):
        """
        looks for a line in topology section 'nonbond_params'
        and modifies it as needed
        :return: None
        """
        subsection = self.top.parameters
        if self.nbs:
            subsection.add_nbfix(type1=self.prefixes[0] + self.types[0], type2=self.prefixes[1] + self.types[1],
                                 mod_sigma=self.dpar, action_default='m')
        elif self.nbe:
            subsection.add_nbfix(type1=self.prefixes[0] + self.types[0], type2=self.prefixes[1] + self.types[1],
                                 mod_epsilon=self.dpar, action_default='m')

    def prefix_type(self, atoms: list, atom_type: str, prefix: str):
        """
        looks for a line in topology section 'atoms'
        and modifies it as needed
        :return: None
        """
        for atom in atoms:
            if "TD-MODDED" not in atom.comment:
                atom.mass_b = atom.mass
                atom.type_b = prefix + atom_type
                atom.charge_b = atom.charge
                atom.comment = atom.comment + " ; TD-MODDED"

    def types_are_consistent(self):
        """
        if many atoms are passed in a single input line, need to make sure
        they are all of the same type; raises ValueError if this is not the case
        :return: None
        """
        for atomset in self.atoms:
            consistent = all([atomset[0].type == atom.type for atom in atomset])
            if not consistent:
                raise ValueError("atoms within a single nbfix modification need to have consistent types")


class ModParam(Mod):
    def __init__(self, td: "gml.ThermoDiff", top: Union[str, gml.Top], structure: str, selections: list, changes: str,
                 period: Optional[int] = -1):
        """
        Subclass to work with modified bonded params (angles and dihedrals).
        The class allows a lists of tuples of atoms that each corresponds
        to a single bonded term, angle or dihedral (all need to have
        the same sets of atomtypes for the calculation to be meaningful);
        the corresponding parameters will be modified in the molecule definition.
        The modification can be applied to all atoms bearing the specific type
        as well as only a subgroup of these.
        :param top: Top, a Top object corresponding to the modified topology
        :param selections: list of str, selections that will define atoms to be modified
        :param changes: str, single character denoting the change to be performed
        """
        super(ModParam, self).__init__(td, top, structure, selections)
        self.dih = 'd' if 'd' in changes else ''  # implement add to top line
        self.ang = 'a' if 'a' in changes else ''  # implement add to top line
        self.period = period
        if self.dih and self.ang:
            raise RuntimeError("Please use separate Mods entries for angles and dihedrals")
        if not (self.dih or self.ang):
            raise ValueError("no mode selected appropriate for an ModParam object")
        if period < 0 and self.dih:
            raise RuntimeError("Please choose periodicity for the dihedral to be modified")
        self.types_are_consistent()
        self.types = [atoms[0].type for atoms in self.atoms]
        self.nums = [[atom.num for atom in atomset] for atomset in self.atoms]
        if self.ang:
            for molname in list({atom.molname for atomset in self.atoms for atom in atomset}):
                self.top.get_molecule(molname).add_ff_params('angles')
                self.mod_ang(molname)
        if self.dih:
            for molname in list({atom.molname for atomset in self.atoms for atom in atomset}):
                self.top.get_molecule(molname).add_ff_params('dihedrals')
                self.mod_dih(molname)

    def __repr__(self):
        """
        string representation of a Mod object, should be unique
        :return: mod one-letter code + Mod index, 1-based
        """
        return "{}-{}, types {}".format(
            ''.join([self.sigma, self.chg, self.eps, self.nbs, self.nbe, self.dih, self.ang]),
            self.counter, self.types)

    def mod_ang(self, molname):
        """
        looks for lines in topology section 'angles'
        and modifies them as needed
        :return: None
        """
        subsection = self.top.get_molecule(molname).get_subsection('angles')
        assert len(self.types) == 3
        for entry in subsection.entries_bonded:
            entry.read_types()
            if all([i == j for i, j in zip(entry.types_state_a, self.types)]) or \
                    all([i == j for i, j in zip(entry.types_state_a, self.types[::-1])]):
                if all([i in numset for i, numset in zip(entry.atom_numbers, self.nums)] or \
                       [i in numset for i, numset in zip(entry.atom_numbers, self.nums[::-1])]):
                    if "TD-MODDED" not in entry.comment:
                        entry.params_state_b = entry.params_state_a[:]
                        entry.params_state_b[0] += self.dpar
                        entry.comment = entry.comment + " ; TD-MODDED"

    def mod_dih(self, molname):
        """
        looks for lines in topology section 'dihedrals'
        and modifies them as needed
        :return: None
        """
        subsection = self.top.get_molecule(molname).get_subsection('dihedrals')
        assert len(self.types) == 4
        for entry in subsection.entries_bonded:
            entry.read_types()
            if all([i == j for i, j in zip(entry.types_state_a, self.types)]) or \
                    all([i == j for i, j in zip(entry.types_state_a, self.types[::-1])]):
                if all([i in numset for i, numset in zip(entry.atom_numbers, self.nums)] or \
                       [i in numset for i, numset in zip(entry.atom_numbers, self.nums[::-1])]):
                    if entry.params_state_a[-1] == self.period:
                        if "TD-MODDED" not in entry.comment:
                            entry.params_state_b = entry.params_state_a[:]
                            entry.params_state_b[1] += self.dpar
                            entry.comment = entry.comment + " ; TD-MODDED"

    def types_are_consistent(self):
        """
        checks whether all input atoms share the same type signature,
        (t1 t2 t3 t4) for dihedrals or (t1 t2 t3) for angles;
        raises ValueError if not the case
        :return: None
        """
        for atomset in self.atoms:
            ok = all([atom.type == atomset[0].type for atom in atomset])
            if not ok:
                raise ValueError("atoms within a single parameter modification need to have consistent types")


class ThermoDiff:
    """
    PLEASE NOTE: certain box types are not suitable for reruns in Gromacs
    (e.g. Amber periodic octahedron). If that's your case, make sure your
    molecules are whole, and use an external tool (e.g. MDTraj) to edit
    your box angles and box side lengths to ensure sufficient space between
    periodic images (PME is not used in reruns anyway, just plain cut-off).
    """
    def __init__(self, name: Optional[str] = None, temperature: Optional[float] = 300):
        self.counter = 0
        self.name = '' if name is None else f'-{name}'
        self.mods = []
        self.temperature = temperature
        self.trajs = []
        self.path = None
        self.derivatives = {}
        self.discrete_free_energy_derivatives = {}
        self.profile_free_energy_derivatives = {}
        self.discrete_observable_derivatives = {}
        self.profile_observable_derivatives = {}
        self.discrete_free_energy_derivative_uncertainties = {}
        self.profile_free_energy_derivative_uncertainties = {}
        self.discrete_observable_derivative_uncertainties = {}
        self.profile_observable_derivative_uncertainties = {}
        self.thresholds = {}

    def add_mod(self, top: Union[str, gml.Top], structure: str, modtype: str, selections: list,
                period: Optional[int] = -1):
        """
        Adds a new modified topology for derivative calculation
        :param top: str or gml.Top, the topology compatible with the system
        :param structure: str, a path to the Pdb or Gro file
        :param modtype: str, type of the modification, one of the following: c, s, e, n, m, a, d
        :param selections: list of str, selections that will define atoms to be modified
        :param period: int, periodicity for a dihedral modification (if applicable)
        :return: None
        """
        if modtype in 'cse':
            self.mods.append(ModAtom(self, top, structure, selections, modtype))
        elif modtype in 'ad':
            self.mods.append(ModParam(self, top, structure, selections, modtype, period))
        elif modtype in 'nm':
            self.mods.append(ModNbfix(self, top, structure, selections, modtype))

    def add_all_nbfix_mods(self, top: Union[str, gml.Top], structure: str, typelist: Optional[list] = None):
        """
        Automatically adds all NBFIX modifications that can be conceived from the parameter set
        :param top: str or gml.Top, the topology compatible with the system
        :param structure: str, a path to the Pdb or Gro file
        :param typelist: optional, only consider the types listed in constructing NBFIXes
        :return: None
        """
        topology = gml.Top(top) if isinstance(top, str) else top
        topology.clear_sections()
        topology.clear_ff_params()
        types = topology.defined_atomtypes if typelist is None \
            else list(set(topology.defined_atomtypes).intersection(set(typelist)))
        types = sorted(types)
        print(f"Finding all combinations between types: {' '.join(types)}")
        for type_pair in sorted(list(combinations_with_replacement(types, 2))):
            print(f"Adding modification: {type_pair[0]}, {type_pair[1]}")
            self.add_mod(deepcopy(topology), structure, modtype='n',
                         selections=[f'type {type_pair[0]}', f'type {type_pair[1]}'])
            self.add_mod(deepcopy(topology), structure, modtype='m',
                         selections=[f'type {type_pair[0]}', f'type {type_pair[1]}'])

    def add_all_sigma_mods(self, top: Union[str, gml.Top], structure: str, exclude: Optional[list] = None,
                           molecules: Optional[Union[str, list]] = None):
        self._add_lj_mods(top, structure, 'sigma', exclude, molecules)

    def add_all_epsilon_mods(self, top: Union[str, gml.Top], structure: str, exclude: Optional[list] = None,
                             molecules: Optional[Union[str, list]] = None):
        self._add_lj_mods(top, structure, 'epsilon', exclude, molecules)

    def add_all_dihedral_mods(self, top: Union[str, gml.Top], structure: str,
                              molecules: Optional[Union[str, list]] = None,
                              selector_type: Optional[Union[str, list]] = None):
        """
        Automatically adds all dihedral modifications that can be conceived from the parameter set
        (one entry at a time, e.g. CA-CT-CT-OH with periodicity 1 produces one modification);
        can be restricted to dihedrals involving selected atom types
        :param top: str or gml.Top, the topology compatible with the system
        :param structure: str, a path to the Pdb or Gro file
        :param molecules: optional list of str, only add dihedrals from these molecules (default is whole system)
        :param selector_type: optional, only consider dihedrals that have this type (if str)
        or any of these types (if list)
        :return: None
        """
        topology = gml.Top(top) if isinstance(top, str) else top
        topology.clear_sections()
        topology.clear_ff_params()
        topology.add_ff_params()
        topology.explicit_defines()
        molecules = [molecules] if isinstance(molecules, str) else molecules
        mols = topology.molecules if molecules is None else [topology.get_molecule(x) for x in molecules]
        dihtypes = []
        for mol in mols:
            try:
                sub = mol.get_subsection('dihedrals')
            except KeyError:
                continue
            for dih_entry in sub.entries_bonded:
                dih_entry.read_types()
                signature = (dih_entry.types_state_a, dih_entry.params_state_a[-1])
                signature_rev = (dih_entry.types_state_a[::-1], dih_entry.params_state_a[-1])
                if signature not in dihtypes and signature_rev not in dihtypes:
                    dihtypes.append(signature)
        if selector_type is not None:
            if isinstance(selector_type, str):
                selector_type = [selector_type]
            dihtypes = [d for d in dihtypes if set(selector_type).intersection(set(d[0]))]
        for n, dih in enumerate(dihtypes):
            sels = [f'type {t}' for t in dih[0]]
            print(f"Adding modification: {', '.join(sels)}, {n}/{len(dihtypes)}")
            self.add_mod(deepcopy(topology), structure, 'd', sels, dih[1])

    def add_all_charge_mods(self, top: Union[str, gml.Top], structure: str,
                            molecules: Optional[Union[str, list]] = None):
        """
        Automatically adds all dihedral modifications that can be conceived from the parameter set
        (one entry at a time, e.g. CA-CT-CT-OH with periodicity 1 produces one modification);
        can be restricted to dihedrals involving selected atom types
        :param top: str or gml.Top, the topology compatible with the system
        :param structure: str, a path to the Pdb or Gro file
        :param molecules: optional list of str, only add dihedrals from these molecules (default is whole system)
        :return: None
        """
        topology = gml.Top(top) if isinstance(top, str) else top
        topology.clear_sections()
        topology.clear_ff_params()
        molecules = [molecules] if isinstance(molecules, str) else molecules
        mols = topology.molecules if molecules is None else [topology.get_molecule(x) for x in molecules]
        mods_list = list(set([(a.atomname, a.resname) for mol in mols for a in mol.atoms]))
        for mod in mods_list:
            print(f"Adding modification: {mod[0]}, {mod[1]}")
            self.add_mod(deepcopy(topology), structure, modtype='c',
                         selections=[f'name {mod[0]} and resname {mod[1]}'])

    def _add_lj_mods(self, top: Union[str, gml.Top], structure: str, which: str, exclude: Optional[list] = None,
                     molecules: Optional[Union[str, list]] = None):
        """
        Automatically adds calculations of derivatives with respect to
        all sigma or epsilon values in the system at hand
        :param top: either str (filename) or gml.Top instance
        :param structure: str, filename of the .pdb or .gro corresponding to the .top file
        :param which: str, 'sigma' or 'epsilon'
        :param exclude: list, optional selection of types to exclude from the calculation
        :param molecules: list or str, optional list of molecules from which the types will be selected
        :return: None
        """
        topology = gml.Top(top) if isinstance(top, str) else top
        molecules = [molecules] if isinstance(molecules, str) else molecules
        include = []
        if molecules is not None:
            for mol in molecules:
                section_mol = topology.get_molecule(mol)
                include.extend(sorted(list({a.type for a in section_mol.atoms})))
        topology.clear_sections()
        topology.clear_ff_params()
        topology.add_ff_params()
        exclude = [] if exclude is None else exclude
        seltypes = sorted(topology.defined_atomtypes) if not include else sorted(include)
        for atomtype in seltypes:
            if atomtype in exclude:
                continue
            assert which in ['sigma', 'epsilon']
            print(f"Adding {which} modification: {atomtype}")
            self.add_mod(deepcopy(topology), structure, modtype=which[0],
                         selections=[f'type {atomtype}'])

    def add_traj(self, top: Union[str, gml.Top], traj: str, datasets: Optional[dict] = None,
                 weights: Optional[list] = None):
        """
        Adds a trajectory on which the derivatives will be calculated,
        including pre-calculated data and frame-wise weights
        :param top: str or gml.Top, the topology compatible with the trajectory
        :param traj: str, path to the trajectory
        :param datasets: dict, named datasets in the format alias: list of [float/int/str]
        :param weights: list of float, per-frame weights (e.g. from Umbrella Sampling, Metadynamics or other reweighing
        schemes)
        :return: None
        """
        if isinstance(top, str):
            top = gml.Top(top)
        datasets = datasets if datasets is not None else {}
        weights = weights if weights is not None else []
        self.trajs.append({'path': traj, 'top': top, 'id': len(self.trajs), 'datasets': datasets, 'weights': weights})

    def get_traj(self, trajid: int) -> dict:
        """
        Gets a trajectory by its integer ID
        :param trajid: int, ID of the trajectory
        :return: dict, the trajectory and its associated data
        """
        return [traj for traj in self.trajs if traj['id'] == trajid][0]

    def add_data_to_traj(self, trajid: int, datasets: dict):
        """
        If not specified in self.add_traj, data can be added to the trajectory
        in a separate step
        :param trajid: int, ID of the trajectory
        :param datasets: dict, named datasets in the format alias: list of [float/int/str]
        :return: None
        """
        traj = self.get_traj(trajid)
        traj['datasets'].update(datasets)

    def add_weights_to_traj(self, trajid: int, weights: list):
        """
        If not specified in self.add_traj, weights can be added to the trajectory
        in a separate step
        :param trajid: int, ID of the trajectory
        :param weights: list of float, per-frame weights (e.g. from Umbrella Sampling, Metadynamics or other reweighing
        schemes)
        :return: None
        """
        traj = self.get_traj(trajid)
        traj['weights'] = weights

    @staticmethod
    def equal_tops(top1: gml.Top, top2: gml.Top) -> bool:
        """
        Checks (in a quick and approximate way) if two topologies are identical
        :param top1: gml.Top, first top to check
        :param top2: gml.Top, second top to check
        :return: bool, whether top1 and top2 are identical
        """
        if all(a1.atomname == a2.atomname for a1, a2 in zip(top1.atoms, top2.atoms)):
            return True
        else:
            return False

    def prep_files(self):
        """
        Prepares the directory tree for the calculations
        :return:
        """
        try:
            os.mkdir(f'working{self.name}')
        except FileExistsError:
            print(f"The 'working{self.name}' directory already exists, will overwrite its contents; clean incomplete "
                  f"'rerun.xvg' files to avoid problems")
        self.path = os.getcwd()
        for mod in self.mods:
            try:
                os.mkdir(f'working{self.name}/{str(mod)}')
            except FileExistsError:
                pass
            mod.save_mod(f'{self.path}/working{self.name}/{str(mod)}', str(mod))

    def launch_reruns(self, parallel, maxcores):
        """
        Launches reruns for each legal mod-trajectory combination
        :param parallel: bool, whether to use parallel calculations
        :param maxcores: int, optional number of processes
        :return: None
        """
        pairs = []
        for mod in self.mods:
            for traj in self.trajs:
                if self.equal_tops(mod.top, traj['top']):
                    datasets = self.get_traj(traj['id'])['datasets']
                    print(f"Calculating rerun for {str(mod)}, traj {traj['path']}")
                    pairs.append((mod, traj, datasets))
        if parallel:
            with ProcessPoolExecutor(maxcores) as p:
                deriv_dicts = p.map(self.launch_rerun, pairs)
        else:
            deriv_dicts = [self.launch_rerun(p) for p in pairs]
        for dd in deriv_dicts:
            self.derivatives.update(dd)

    @staticmethod
    def launch_rerun(mod_traj_ds):
        """
        Launches an individual alchemical rerun and reads the data, can be parallelized
        (skips the calculation if already performed)
        :param mod_traj_ds: tuple, contains the mod, traj and dataset objects
        :return: None
        """
        derivatives = {}
        mod, traj, datasets = mod_traj_ds
        mod.goto_mydir()
        if 'rerun.xvg' not in os.listdir('.'):
            derivatives[(mod.counter, traj['id'])] = gml.calc_gmx_dhdl('../../' + mod.structure, str(mod) + '-' +
                                                                       mod.top.top, '../../' + traj['path'], nb='cpu',
                                                                       pme='cpu')
        else:
            derivatives[(mod.counter, traj['id'])] = gml.read_xvg('rerun.xvg', [0])
        for key in datasets.keys():
            if not len(derivatives[(mod.counter, traj['id'])]) == len(datasets[key]):
                raise RuntimeError(f"The number of points in dataset {key} for trajectory num {traj['id']} "
                                   f"({len(datasets[key])}) does not match the derivatives "
                                   f"({len(derivatives[(mod.counter, traj['id'])])})")
        os.chdir('..')
        return derivatives

    def calc_weights(self):
        """
        Normalizes weights (or adds constant weights if not specified)
        :return: None
        """
        for traj in self.trajs:
            if len(traj['weights']) == 0:
                traj['weights'] = [1] * self.get_length(traj['id'])
            wsum = sum(traj['weights'])
            traj['weights'] = [w / wsum for w in traj['weights']]

    def run(self, parallel=True, maxcores=None):
        """
        Runs the set of calculations and prepares everything
        for data extraction
        :return: None
        """
        self.prep_files()
        self.launch_reruns(parallel, maxcores)
        self.calc_weights()

    def get_mod(self, counter: int):
        """
        Gets the desired gml.Mod subclass instance by counter (ID)
        :param counter: int, ID of the modified topology
        :return: gml.Mod, selected modified topology
        """
        return [mod for mod in self.mods if mod.counter == counter][0]

    def get_length(self, trajid):
        """
        Gets the length of the trajectory from the length
        of the derivatives file (provided reruns have been performed)
        :param trajid: int, ID of the trajectory
        :return: int, length of the trajectory
        """
        for key in self.derivatives.keys():
            if key[1] == trajid:
                return len(self.derivatives[key])

    def dataset_is_numeric(self, dataset: str):
        """
        Checks if the user-specified dataset is numeric or string-based
        (in the latter case discrete states will be used)
        :param dataset: str, alias of the dataset
        :return: bool, whether the dataset is numeric
        """
        flat_dataset = {data for traj in self.trajs for data in traj['datasets'][dataset]}
        if all([isinstance(x, str) for x in flat_dataset]):
            return False
        elif all([isinstance(x, int) or isinstance(x, float) for x in flat_dataset]):
            return True
        else:
            raise RuntimeError(f"Dataset {dataset} contains mixed numeric and non-numeric entries")

    def get_flat_data(self, binning_dataset: str, deriv_dataset: str, mod) -> Tuple[list, list, list, list]:
        """
        Selects all numeric datasets that correspond to a given alias
        and provides flat (1-D) lists for data, weights, and derivatives
        :param binning_dataset: str, alias of the dataset used for binning/thresholding
        :param deriv_dataset: str, alias of the dataset used for derivative calculation (can be same as binning_dataset)
        :param mod: gml.Mod
        :return: tuple of lists: data, weights, and derivatives
        """
        counters_trajids = list(self.derivatives.keys())
        active_dataset_trajids = [traj['id'] for traj in self.trajs if binning_dataset in traj['datasets'].keys()
                                  and deriv_dataset in traj['datasets'].keys()]
        active_counters_trajids = [ct for ct in counters_trajids if ct[1] in active_dataset_trajids
                                   and ct[0] == mod.counter]
        flat_bin_data = [data for ct in active_counters_trajids for data in
                         self.get_traj(ct[1])['datasets'][binning_dataset]] # ids of the bins
        flat_deriv_data = [data for ct in active_counters_trajids for data in
                           self.get_traj(ct[1])['datasets'][deriv_dataset]] # the property
        flat_weights = [wght for ct in active_counters_trajids for wght in self.get_traj(ct[1])['weights']] # normalized
        flat_derivs = [der for ct in active_counters_trajids for der in self.derivatives[ct]] # the dU/d(sigma) values
        return flat_bin_data, flat_deriv_data, flat_weights, flat_derivs

    def calc_discrete_derivatives(self, dataset: Optional[str] = None, free_energy: Optional[bool] = True,
                                  threshold: Optional[list] = None, cv_dataset: Optional[str] = None,
                                  noe: Optional[bool] = True, bootstrap = False):
        """
        Calculates selected derivatives for a number of discrete states
        :param dataset: str, alias for the dataset
        :param free_energy: bool, whether the derivative should be of free energy or of observable/CV
        :param threshold: list, for continues dataset specifies 2N boundaries used to define N discrete states
        :param cv_dataset: str, if specified then derivatives will be calculated based on 'dataset' but binning will be
        :param noe: bool, if free_energy is False, then this switches on calculations based on averaging the -6th power (use one state)
        :param bootstrap: bool, whether to estimate uncertainty using bootstrap
        performed based on 'cv_dataset'
        :return: None
        """
        if dataset is None:
            print("No dataset label chosen, will assume each trajectory defines a separate state")
            for traj in self.trajs:
                traj['datasets']['discrete'] = [str(traj['id']) for _ in range(self.get_length(traj['id']))]
            dataset = 'discrete'
        if self.dataset_is_numeric(dataset) and threshold is None:
            raise RuntimeError("When calculating a derivative with numeric datasets, specify a threshold; when "
                               "calculating a derivative over discrete states, use strings as state labels")
        if threshold is not None:
            if len(threshold) % 2 == 1:
                raise RuntimeError("The list 'threshold' has to have an even number of entries, one for each starting "
                                   "and ending point of each state")
        states = [(round(x, 6), round(y, 6)) for x, y in
                  zip(threshold[::2], threshold[1::2])] if threshold is not None else \
            sorted(list({data for traj in self.trajs for data in traj['datasets'][dataset]}))
        binning_dset, deriv_dset = (dataset, dataset) if cv_dataset is None else (cv_dataset, dataset)
        self.thresholds[dataset] = threshold
        for mod in self.mods:
            binning_data, deriv_data, weights, derivs = self.get_flat_data(binning_dset, deriv_dset, mod)
            vals_derivatives = {st: [] for st in states + [None]}
            vals_data = {st: [] for st in states + [None]}
            vals_weights = {st: [] for st in states + [None]}
            mean_derivatives = {st: 0 for st in states + [None]}
            mean_product = {st: 0 for st in states + [None]}
            mean_data = {st: 0 for st in states + [None]}
            unc_derivatives = {st: 0 for st in states + [None]}
            unc_product = {st: 0 for st in states + [None]}
            unc_data = {st: 0 for st in states + [None]}
            counter = {st: 0 for st in states + [None]}
            if noe and not free_energy:
                deriv_data = [x ** (-6) for x in deriv_data]
            # first preparing/sorting values into different states
            for n in range(len(binning_data)):
                state_index = None
                if threshold is not None:
                    for x, y in zip(threshold[::2], threshold[1::2]):
                        if x <= binning_data[n] <= y:
                            state_index = (round(x, 6), round(y, 6))
                            break
                else:
                    state_index = binning_data[n]
                vals_derivatives[state_index].append(derivs[n])
                vals_data[state_index].append(deriv_data[n])
                vals_weights[state_index].append(weights[n])
                counter[state_index] += weights[n]
            # then calculating the weighted means of all components
            for state_index in counter.keys():
                if counter[state_index] > 0:
                    mean_derivatives[state_index] = (sum([w*dr for w, dr
                                                         in zip(vals_weights[state_index],
                                                                vals_derivatives[state_index])]) /
                                                     (counter[state_index] * mod.dpar))
                    mean_product[state_index] = (sum([w*d*dr for w, d, dr
                                                     in zip(vals_weights[state_index],
                                                            vals_data[state_index],
                                                            vals_derivatives[state_index])]) /
                                                 (counter[state_index] * mod.dpar))
                    mean_data[state_index] = (sum([w*d for w, d
                                                  in zip(vals_weights[state_index],
                                                         vals_data[state_index])]) /
                                              counter[state_index])
                    # finally, if needed, computing the uncertainties with bootstrap
                    if bootstrap:
                        unc_derivatives[state_index] = self.bootstrap(vals_derivatives[state_index],
                                                                      weights=vals_weights[state_index]) / mod.dpar
                        unc_data[state_index] = self.bootstrap(vals_data[state_index],
                                                               weights=vals_weights[state_index])
                        unc_product[state_index] = self.bootstrap([d*dr for d, dr
                                                                   in zip(vals_data[state_index],
                                                                          vals_derivatives[state_index])],
                                                                  weights=vals_weights[state_index]) / mod.dpar
            # here just writing the results to the final dictionaries
            if free_energy:
                self.discrete_free_energy_derivatives[(str(mod), dataset)] = [mean_derivatives[x]
                                                                              for x in mean_derivatives.keys()
                                                                              if counter[x] > 0]
                if bootstrap:
                    self.discrete_free_energy_derivative_uncertainties[(str(mod), dataset)] = [unc_derivatives[x]
                                                                                               for x in unc_derivatives.keys()
                                                                                               if counter[x] > 0]
            else:
                mean_obs = {key: 0 for key in mean_derivatives.keys()}
                final_unc = {key: 0 for key in mean_derivatives.keys()}
                for key in mean_derivatives.keys():
                    mean_obs_der = (1 / (0.008314 * self.temperature)) * (mean_data[key] * mean_derivatives[key] -
                                                                          mean_product[key])
                    if not noe:
                        mean_obs[key] = mean_obs_der
                    else:
                        mdk = mean_data[key] ** (-7/6) if mean_data[key] != 0 else 0
                        mean_obs[key] = -1/6 * mdk * mean_obs_der
                    if bootstrap:
                        if counter[key] > 0:
                            rel_data = unc_data[key]/mean_data[key]
                            rel_der = unc_derivatives[key]/mean_derivatives[key]
                            unc_obs = np.sqrt(unc_product[key] ** 2 +
                                              (mean_data[key] * mean_derivatives[key] * (rel_der + rel_data))**2)
                            if not noe:
                                final_unc[key] = unc_obs
                            else:
                                final_unc[key] = mean_obs[key] * (unc_obs/mean_obs_der +
                                                                  7/6 * mean_data[key]**(-13/6) * rel_data)
                self.discrete_observable_derivatives[(str(mod), dataset)] = [mean_obs[x] for x in mean_obs.keys()
                                                                             if counter[x] > 0]
                if bootstrap:
                    self.discrete_observable_derivative_uncertainties[(str(mod), dataset)] = [final_unc[x] for x
                                                                                              in final_unc.keys()
                                                                                              if counter[x] > 0]

    def bootstrap(self, orig_sample, n_resamples=100, weights=None):
        import numpy as np
        sample = np.array(orig_sample)
        weights = np.ones(len(sample))/len(sample) if weights is None else weights/np.sum(weights)
        means = []
        for i in range(n_resamples):
            new_sample = np.random.choice(sample, size=len(sample), replace=True, p=weights)
            means.append(np.mean(new_sample))
        return np.std(means)

    def calc_profile_derivatives(self, dataset: str, free_energy: Optional[bool] = True, nbins: Optional[int] = 50,
                                 cv_dataset: Optional[str] = None):
        """
        Calculates selected derivatives of a profile
        :param dataset: str, alias for the dataset
        :param free_energy: bool, whether the derivative should be of free energy or of observable/CV
        :param nbins: int, number of bins that will be used to calculate the profile
        :param cv_dataset: str, if specified then derivatives will be calculated based on 'dataset' but binning will be
        performed based on 'cv_dataset'
        :return: None
        """
        binning_dset, deriv_dset = (dataset, dataset) if cv_dataset is None else (cv_dataset, dataset)
        for mod in self.mods:
            binning_data, deriv_data, weights, derivs = self.get_flat_data(binning_dset, deriv_dset, mod)
            dmin, dmax = min(binning_data), max(binning_data)
            thresh = [dmin + n * (dmax - dmin) / (nbins + 1) for n in range(nbins + 1)]
            mean_derivatives = {0.5 * (x + y): 0 for x, y in zip(thresh[:-1], thresh[1:])}
            mean_product = {0.5 * (x + y): 0 for x, y in zip(thresh[:-1], thresh[1:])}
            mean_data = {0.5 * (x + y): 0 for x, y in zip(thresh[:-1], thresh[1:])}
            for n in range(len(binning_data)):
                for x, y in zip(thresh[:-1], thresh[1:]):
                    if x <= binning_data[n] < y:
                        mean_derivatives[0.5 * (x + y)] += weights[n] * derivs[n]
                        if not free_energy:
                            mean_product[0.5 * (x + y)] += deriv_data[n] * weights[n] * derivs[n]
                            mean_data[0.5 * (x + y)] += deriv_data[n]
            if free_energy:
                self.profile_free_energy_derivatives[(str(mod), dataset)] = ([(x + y) / 2 for x, y in zip(thresh[:-1],
                                                                                                          thresh[1:])],
                                                                             [mean_derivatives[x] / mod.dpar
                                                                              for x in mean_derivatives.keys()])
            else:
                mean_obs = {key: 0 for key in mean_derivatives.keys()}
                for key in mean_derivatives.keys():
                    mean_obs[key] = (1 / 0.008314 * self.temperature) * (
                            mean_data[key] * mean_derivatives[key] - mean_product[key])
                self.profile_observable_derivatives[(str(mod), dataset)] = ([(x + y) / 2 for x, y in zip(thresh[:-1],
                                                                                                         thresh[1:])],
                                                                            [mean_obs[x] / mod.dpar
                                                                             for x in mean_obs.keys()])

    def write_discrete_derivatives(self, dataset: str, free_energy: bool, bootstrap: bool = False):
        """
        Writes the derivatives to individual files in each working subdirectory
        :param dataset:
        :param free_energy:
        :param bootstrap:
        :return:
        """
        derivs = self.discrete_free_energy_derivatives if free_energy else self.discrete_observable_derivatives
        for key in derivs.keys():
            if key[1] == dataset:
                mod = key[0]
                with open(f'working{self.name}/{mod}/{key[1]}-discrete_sensitivity.dat', 'w') as outfile:
                    outfile.write(f"{round(derivs[key][1] - derivs[key][0], 3)}\n")

    def print_discrete_derivatives(self, dataset: str, free_energy: bool, outfile: Optional[str] = None, bootstrap: bool = False):
        """
        Prints discrete derivatives to the screen or a file
        :param dataset: str, for which dataset the derivatives should be selected
        :param free_energy: bool, whether the 1st state should be set to reference (equal to 0.0)
        :param outfile: str, optional filename if the result is to be printed to a file
        :param bootstrap: bool, whether to add uncertainty
        :return: None
        """
        if outfile is not None:
            outfile = open(outfile, mode='w')
        derivs = self.discrete_free_energy_derivatives if free_energy else self.discrete_observable_derivatives
        uncerts = self.discrete_free_energy_derivative_uncertainties if free_energy else self.discrete_observable_derivative_uncertainties
        thrs = self.thresholds[dataset]
        for x in range(len(thrs)//2):
            x1 = round(thrs[2 * x], 3)
            x2 = round(thrs[2 * x + 1], 3)
            xx = f"{x1}-{x2}"
            print(f'  {xx:20s}', end='|', file=outfile)
        print(f'{"  others":20s}  |', file=outfile)
        print(23 * (len(thrs)//2 + 1) * '-', file=outfile)
        all_ders = [q for q in derivs.keys() if q[1] == dataset]
        if free_energy:
            sorted_ders = sorted(all_ders, key=lambda l: abs(derivs[l][1] - derivs[l][0]), reverse=True)
        else:
            sorted_ders = sorted(all_ders, key=lambda l: abs(derivs[l][0]), reverse=True)
        strings = {}
        for n, mod in enumerate(all_ders):
            strings[mod] = []
            for nd, d in enumerate(derivs[mod]):
                if bootstrap:
                    err = f' +/- {uncerts[mod][nd]}'
                else:
                    err = ''
                if free_energy:
                    strings[mod].append(f'  {d-derivs[mod][0]:20.5f}{err}|')
                else:
                    strings[mod].append(f'  {d:20.5f}{err}|')
            try:
                strings[mod].append(f"  {str(self.mods[n]):6s}  {'-'.join(self.mods[n].types):16s}  {self.mods[n].period:4d}\n")
            except:
                try:
                    strings[mod].append(f"  {str(self.mods[n]):6s}  {'-'.join(self.mods[n].types):16s}\n")
                except:
                    if str(self.mods[n]).startswith('c'):
                        strings[mod].append(f"  {str(self.mods[n]):6s}  {self.mods[n].atoms[0].atomname:6s}  {self.mods[n].atoms[0].resname:6s}\n")
                    else:
                        strings[mod].append(f"  {str(self.mods[n]):6s}  {self.mods[n].type:16s}\n")
            # strings[mod].append(23 * (len(thrs)//2 + 1) * '-')
        for sormod in sorted_ders:
            print(''.join(strings[sormod]), file=outfile)
        if outfile is not None:
            outfile.close()



import gromologist as gml
from copy import deepcopy
import os
from glob import glob
from subprocess import call
from multiprocessing import Pool

try:
    from scipy.optimize import dual_annealing
except ImportError:
    pass
try:
    import numpy as np
except ImportError:
    pass


class DihOpt:
    def __init__(self, top, qm_ref, cutoff=None, qm_unit='kj/mol', processes=1, tmpi=False, **kwargs):
        """
        The DihOpt is a dihedral optimizer that reads in (preferably) a Gaussian .log
        file and a Gromacs topology with selected dihedral terms marked as DIHOPT in the comment field,
        and then optimizes the dihedral parameters to minimize the RMSE between the QM and MM energies.
        :param top: a gml.Top object or str, Gromacs topology (either read-in or filename)
        :param qm_ref: str or list of str, path(s) to Gaussian .log files
        :param cutoff: float, will not attempt to fit data points above this threshold
        :param qm_unit: str, if qm_ref is a list of energies, this should be 'kcal/mol', 'kj/mol' or 'hartree'
        :param processes: int, how many independent fitting procedures should be launched (supports parallelism)
        :param tmpi: bool, whether to use Gromacs with thread-MPI in parallel execution
        :param kwargs: dict, if 'top' is a path, this will be passed to the gml.Top constructor
        """
        if not isinstance(qm_ref, list):
            qm_ref = [qm_ref]
        if isinstance(top, list):
            if isinstance(top[0], str):
                self.orig_top = [gml.Top(t, **kwargs) for t in top]
            else:
                self.orig_top = top
        else:
            if isinstance(top, gml.Top):
                self.orig_top = [top for _ in qm_ref]
            else:
                self.orig_top = [gml.Top(top, **kwargs) for _ in qm_ref]
        self.orig_vals = None
        self.best_rmse = None
        self.best_top = self.orig_top
        self.optimizable_params = self.get_opt_dih(self.orig_top)
        self.traj = []
        self.frame = []
        self.lowest_iter_rmse = 1000000
        self.curr_iter = 0
        self.opt_results = None
        self.print_progress = False
        self.all_opts = None
        self.energy_term = 0
        self.energy_profiles_opt = []
        self.check_imports()
        self.check_dihopt()
        # TODO so far only works with Gaussian outputs
        qm_unit = qm_unit.lower()
        if qm_unit not in ['kj/mol', 'kcal/mol', 'hartree']:
            raise RuntimeError("Unit can be kj/mol, kcal/mol or hartree")
        # we're converting everything to kJ/mol
        scaling = 1.0 if qm_unit == 'kj/mol' else 4.184 if 'kcal/mol' else 2625.5
        self.cutoff = cutoff * scaling if cutoff is not None else 10000000
        if isinstance(qm_ref, list) and type(qm_ref[0]) == str:  # a list of Gaussian logs
            self.qm_ref = [self.read_gaussian_energies(ref, n) for n, ref in enumerate(qm_ref)]
        else:
            raise RuntimeError("As qm_ref, pass a Gaussian .log file, or a list of Gaussian .logs.")
        assert len(self.orig_top) == len(self.qm_ref)
        self.qm_ref = [[x - min(ref) for x in ref] for ref in self.qm_ref]
        self.mod_tops = [[deepcopy(top) for top in self.orig_top] for _ in range(processes)]
        self.processes = processes
        self.run_count = 0
        self.tmpi = tmpi
        self.gmx = self.find_gmx()
        self.gen_dirs()
        self.write_mdp()
        self.opt_prof = None

    def check_imports(self):
        """
        To keep the lib as lightweight as possible, we're not explicitly requiring scipy/numpy,
        but to use DihOpt both are required, so here we look for them before running
        :return: None
        """
        try:
            _ = dual_annealing
        except NameError:
            raise ImportError("scipy needed to run dihedral optimization, run 'pip install scipy' in command line")
        try:
            _ = np.array
        except NameError:
            raise ImportError("numpy needed to run dihedral optimization, run 'pip install numpy' in command line")

    def check_dihopt(self):
        if len(self.orig_top) > 1:
            first = self.orig_top[0].parameters.get_opt_dih(types=True)
            if not all(all(ff[i] == tt[i] for i in range(5)) for t in self.orig_top[1:]
                       for ff, tt in zip(first, t.parameters.get_opt_dih(types=True))):
                raise RuntimeError("Not all .top files have the same order and quantity of dihedral parameters set for"
                                   " optimization with the DIHOPT keyword; look up the documentation")

    @staticmethod
    def get_opt_dih(tops):
        return tops[0].parameters.get_opt_dih()

    @staticmethod
    def set_opt_dih(tops, ff_values):
        for top in tops:
            top.parameters.set_opt_dih(ff_values)

    def calc_energy(self, ff_values=None, sys=0, cleanup=True):
        """
        Calculates the energy profile for a given set of dihedral parameters
        by running a rerun and extracting data from the resulting .xvg file
        :param ff_values: list of floats, dihedral parameters that will be set for the current iteration
        :param sys: int, index of the system
        :param cleanup: bool, whether to remove all files after the iteration
        :return: np.array, classical (MM) potential energies in kJ/mol for each frame in the trajectory
        """
        ff_values = self.get_opt_dih(self.best_top) if ff_values is None else ff_values
        self.set_opt_dih(self.mod_tops[sys], ff_values)
        energy = []
        for n, top in enumerate(self.mod_tops[sys]):
            top.save_top('opt{}/mod{}.top'.format(sys, n))
            self.gmx_grompp(sys, n)
            self.gmx_mdrun(sys, n)
            energy.append(self.gmx_energy(sys, n))
            if cleanup:
                self.cleanup(sys, n)
        return np.concatenate(energy)

    def calc_rmse(self, ff_values=None, sys=0, cleanup=True):
        """
        For a given set of dihedral parameters, runs the rerun and calculates the RMSE
        between the QM data and the new MM energies
        :param ff_values: list of floats, dihedral parameters that will be set for the current iteration
        :param sys: int, index of the system
        :param cleanup: bool, whether to remove all files after the iteration
        :return: float, root-mean-square-error (in kJ/mol) between the QM and MM potential energy profiles
        """
        ff_values = self.get_opt_dih(self.best_top) if ff_values is None else ff_values
        energies = self.calc_energy(ff_values, sys, cleanup)
        rmse = [(x-y)**2 for x, y in zip(energies, [v for val in self.qm_ref for v in val]) if y < self.cutoff]
        rmse = (sum(rmse)/len(rmse))**0.5
        call('echo {r} >> opt{s}/rmse'.format(s=sys, r=rmse), shell=True)
        if self.curr_iter % 10 == 0 and int(self.curr_iter) > 0 and self.print_progress:
            print("Iteration {} in process {}, current lowest RMSE is {}".format(int(self.curr_iter), sys,
                                                                                 self.lowest_iter_rmse))
            self.print_progress = False
        if abs(self.curr_iter % 10) > 0 and not self.print_progress:
            self.print_progress = True
        return rmse

    def get_bounds(self, sys=0):
        """
        Finds bounds for parameter optimization, odd ones are for angle offsets,
        even ones for force constants
        :param sys: int, index of the system
        :return: list of tuples, bounds for the optimization procedure
        """
        vals = self.get_opt_dih(self.mod_tops[sys])
        return [(0, 360), (0, 15)] * (len(vals) // 2)

    def optimize(self, maxiter=200):
        """
        Calculates initial values, then performs the actual optimization
        using sklearn.optimize.dual_annealing either in a single process
        (when self.processes is 1) or in parallel (when self.processes > 1)
        :param maxiter: int, at how many iterations to terminate
        :return: None
        """
        # check for imports first
        self.run_count += 1
        self.orig_vals = self.calc_energy() if self.orig_vals is None else self.orig_vals
        self.best_rmse = self.calc_rmse()
        print("Running optimization in {} process{}, initial RMSE "
              "is {}".format(self.processes, 'es' if self.processes > 1 else '', self.best_rmse))
        if self.processes == 1:
            self.opt_results = dual_annealing(self.calc_rmse, bounds=self.get_bounds(0), maxiter=maxiter,
                                              x0=self.get_opt_dih(self.best_top), callback=self.progress)
            # opt_vals = shgo(self.calc_rmse, args=(0,), bounds=self.get_bounds(0), options={'maxev': 1000})
            self.best_top = self.mod_tops[0]
            print("\nOptimized parameters:\n")
            print("    RMSE:       {:10.3f}".format(self.opt_results.fun))
            print("    Parameters: {}\n".format([round(x, 6) for x in self.opt_results.x]))
            print("Initial parameters:\n")
            print("    RMSE:       {:10.3f}".format(self.best_rmse))
            print("    Parameters: {}\n".format([round(x, 6) for x in self.get_opt_dih(self.orig_top)]))
        else:
            p = Pool()
            self.all_opts = p.map(mappable, [(i, self, maxiter) for i in range(self.processes)])
            best_model_index = int(np.argmin([q.fun for q in self.all_opts]))
            self.opt_results = self.all_opts[best_model_index]
            self.best_top = self.mod_tops[best_model_index]
            print("\nObtained {} models with the following RMSE values:\n".format(len(self.all_opts)))
            for i in range(len(self.all_opts)):
                print("    Model {}".format(i))
                print("    RMSE:       {:10.3f}".format(self.all_opts[i].fun))
                print("    Parameters: {}\n".format([round(x, 6) for x in self.all_opts[i].x]))
            print("Initial parameters:\n")
            print("    RMSE:       {:10.3f}".format(self.best_rmse))
            print("    Parameters: {}\n".format([round(x, 6) for x in self.get_opt_dih(self.orig_top)]))
            print("Choosing the lowest-RMSE model {}".format(best_model_index))
        self.set_opt_dih(self.best_top, self.opt_results.x)
        for n, top in enumerate(self.best_top):
            top.save_top('opt{}_topol{}.top'.format(self.run_count, n))
        self.energy_profiles_opt.append(self.calc_energy(self.opt_results.x))

    def restart(self, maxiter=200):
        """
        Can be used to reoptimize the parameters, using previous
        best result as the new starting point
        :param maxiter: int, at how many iterations to terminate
        :return:
        """
        self.best_top = [gml.Top(i) for i in glob('opt{}_topol[0-9]*.top'.format(self.run_count))]
        self.optimize(maxiter=maxiter)

    def gmx_grompp(self, sys=0, n=0):
        """
        Runs gmx grompp to prepare a .tpr file for the rerun
        :param sys: int, index of the system
        :param n: int, index of the topology
        :return:
        """
        call('{gmx} grompp -f run.mdp -p opt{sys}/mod{n}.top -c {frame} -o opt{sys}/mod{n}.tpr -po opt{sys}/mdout.mdp '
             '-maxwarn 5 >> gmp.log 2>&1'.format(gmx=self.gmx, sys=sys, frame=self.frame[n], n=n), shell=True)

    def gmx_mdrun(self, sys=0, n=0):
        """
        Runs gmx mdrun with the '-rerun' option to extract MM energies
        :param sys: int, index of the system
        :param n: int, index of the topology
        :return:
        """
        if self.tmpi:
            tmpi = ' -ntmpi 1 '
        else:
            tmpi = ''
        call('{gmx} mdrun -s opt{sys}/mod{n}.tpr -o opt{sys}/mod{n}.trr -x opt{sys}/mod{n}.xtc -e opt{sys}/mod{n}.edr '
             '-g opt{sys}/mod{n}.log -c opt{sys}/mod{n}.gro -v -ntomp 1 {tmpi} -rerun {traj} -nb cpu >> mdr{n}.log '
             '2>&1'.format(gmx=self.gmx, sys=sys, tmpi=tmpi, traj=self.traj[n], n=n), shell=True)

    def gmx_energy(self, sys=0, n=0):
        """  # TODO merge all
        Runs gmx energy on the .edr file produced by the rerun, also
        extracts energies (in kJ/mol) from it & shifts them to 0
        :param sys: int, index of the system
        :param n: int, index of the topology
        :return: numpy.array, energies of individual frames in kJ/mol
        """
        if not self.energy_term:
            self.get_energy_term(sys)
        ge = '{} energy'.format(self.gmx) if self.gmx else 'g_energy'
        call('echo "{et} 0" | {ge} -f opt{sys}/mod{n}.edr -o opt{sys}/mod{n}.xvg >> edr{n}.log '
             '2>&1'.format(ge=ge, sys=sys, et=self.energy_term, n=n), shell=True)
        vals = np.loadtxt('opt{sys}/mod{n}.xvg'.format(sys=sys, n=n), comments=['#', '@'])[:, 1]
        return vals - np.min(vals)

    def get_energy_term(self, sys=0, n=0):
        """
        Checks the index of the 'Potential' term in gmx energy
        (usually 8 or 9, depending on the presence of impropers)
        :param sys: int, index of the system
        :param n: int, index of the topology
        :return: None
        """
        ge = '{} energy'.format(self.gmx) if self.gmx else 'g_energy'
        call('echo "0" | {ge} -f opt{sys}/mod{n}.edr > legend.log 2>&1'.format(ge=ge, sys=sys, n=n), shell=True)
        self.energy_term = os.popen('cat legend.log | grep Potential | awk \'{for (i=1;i<=NF;i++){if ($i=="Potential")'
                                    ' {print $(i-1)}}}\'').read().strip()

    @staticmethod
    def cleanup(sys=0, n=0):
        """
        Removes files on-the-fly to prevent errors at 99 backups
        :param sys: int, index of the system
        :param n: int, index of the topology
        :return: None
        """
        call('rm opt{s}/mdout.mdp opt{s}/mod{n}.xvg opt{s}/mod{n}.edr opt{s}/mod{n}.tpr '
             'opt{s}/mod{n}.log'.format(s=sys, n=n), shell=True)

    def plot_fit(self):
        """
        If optimization is complete, plots the resulting energy profile against
        both original MM energies and the QM reference
        :return: None
        """
        import matplotlib.pyplot as plt
        if self.opt_results is not None:
            plt.clf()
            plt.plot(self.orig_vals, label='original MM')
            for n, array in enumerate(self.energy_profiles_opt, 1):
                plt.plot(array, label='optimized MM (round {})'.format(n))
            plt.plot(np.concatenate(self.qm_ref), label='QM reference', c='k')
            plt.legend()
            plt.xlabel('trajectory frame')
            plt.ylabel('energy [kJ/mol]')
            plt.show()

    def progress(self, _, f, context):
        """
        Constantly updates the curr_iter and lowest_iter_rmse attributes used for reporting
        :param _: discarded
        :param f: float, current function value (RMSE)
        :param context: int, key corresponding to different outcomes in the optimization algorithm
        :return: None
        """
        if context in [0, 1, 2]:
            self.curr_iter += 1.0
            if f < self.lowest_iter_rmse:
                self.lowest_iter_rmse = f

    @staticmethod
    def write_mdp():
        """
        Writes an .mdp file with all options necessary for a rerun
        :return: None
        """
        mdp_defaults = {"nstenergy": 1, "nstlog": 0, "nstcalcenergy": 1, "nstxout-compressed": 0, "pbc": "xyz",
                        "coulombtype": "Cut-off", "vdw-type": "Cut-off", "rlist": 2.0, "rcoulomb": 2.0, "rvdw": 2.0}
        mdp = '\n'.join(["{} = {}".format(param, value) for param, value in mdp_defaults.items()])
        with open('run.mdp', 'w') as outfile:
            outfile.write(mdp)

    def gen_dirs(self):
        """
        Makes separate directories for all the subprocesses
        :return: None
        """
        for i in range(self.processes):
            try:
                os.mkdir('opt{}'.format(i))
            except FileExistsError:
                pass

    @staticmethod
    def find_gmx():
        """
        Attempts to find Gromacs executables to run gmx grompp/mdrun/energy
        :return: str, path to the gmx/gmx_mpi/gmx_d executable
        """
        gmx = os.popen('which gmx 2> /dev/null').read().strip()
        if not gmx:
            gmx = os.popen('which gmx_mpi 2> /dev/null').read().strip()
        if not gmx:
            gmx = os.popen('which gmx_d 2> /dev/null').read().strip()
        if not gmx:
            gmx = os.popen('which grompp 2> /dev/null').read().strip()
            if gmx:
                gmx = ''
            else:
                raise RuntimeError("No gromacs executable (gmx, gmx_mpi, gmx_d, grompp) found in $PATH")
        return gmx

    def read_gaussian_energies(self, log, n_sys, traj='opt_traj', structfile='struct'):
        """
        Function that parses a Gaussian .log file with a dihedral scan results,
        extracting optimized geometries and reference QM energies; also saves
        the trajectory/structure as .pdb files based on .top atom names
        :param log: str, name of the Gaussian .log file
        :param n_sys: int, index of the log file (in case there are many)
        :param traj: str, name of the output .pdb trajectory file
        :param structfile: str, name of the output .pdb structure file needed for gmx grompp
        :return: list, energies in kJ/mol
        """
        log_contents = [x.strip().split() for x in open(log)]
        natoms = [int(x[1]) for x in log_contents if x and x[0] == "NAtoms="][0]
        energies = []
        read_geo = False
        geoms = []
        energy_tmp = None
        for n, line in enumerate(log_contents):
            if len(line) > 2 and line[0] == 'SCF' and line[1] == 'Done:':
                energy_tmp = float(line[4])
            if len(line) > 3 and line[1] == 'Stationary' and line[2] == 'point' and line[3] == 'found.':
                if energy_tmp is not None:
                    energies.append(energy_tmp)
                else:
                    raise RuntimeError("Keyword 'Stationary point found' before first 'SCF done' in Gaussian log, check"
                                       " your QM run")
                read_geo = True
            if read_geo and len(line) > 1 and line[0] == 'Standard' and line[1] == 'orientation:':
                struct = np.array([[float(line[x]) for x in [3, 4, 5]] for line in log_contents[n+5:n+5+natoms]])
                geoms.append(struct)
                read_geo = False
        opt_traj = gml.Traj(top=self.orig_top[n_sys], array=geoms)  # TODO make list of trajs
        opt_traj.save_traj_as_pdb('{}_{}.pdb'.format(traj, n_sys))
        opt_traj.save_pdb('{}_{}.pdb'.format(structfile, n_sys))
        self.traj.append('{}_{}.pdb'.format(traj, n_sys))
        self.frame.append('{}_{}.pdb'.format(structfile, n_sys))
        return [2625.5 * x for x in energies]

    def make_movie(self):
        """
        Uses Molywood to produce a movie showing the optimization results
        along with the QM optimization trajectory, highlighting the dihedral
        being optimized
        :return: None
        """
        dihedrals_indices = self.orig_top.parameters.get_opt_dih_indices()
        np.savetxt('energies.dat', np.vstack([np.linspace(0, len(self.orig_vals), len(self.orig_vals))] + [self.qm_ref]
                                             + [self.orig_vals] + self.energy_profiles_opt).T)
        moly_inp = "$ global fps={} name=dihopt\n$ scene structure={} resolution=700,700" \
                   "\n\n#scene\n".format(len(self.orig_vals)/8, self.traj)
        moly_inp = moly_inp + 'highlight selection="all" style=licorice color=type mode=u\n'
        for dih in dihedrals_indices:
            moly_inp = moly_inp + 'highlight selection="serial {} {} {} {}" style=licorice thickness=1.1 ' \
                                  'alpha=0.65 mode=u color=green\n'.format(*dih)
        moly_inp = moly_inp + 'insert_tcl code="display resetview"\nzoom_out scale=1.5\nanimate frames=0\n'
        moly_inp = moly_inp + 'do_nothing t=2s\n'
        moly_inp = moly_inp + '{{animate t=8s frames=0:{}; add_overlay datafile=energies.dat origin=0.65,0.65 ' \
                              'relative_size=0.35}}\n'.format(len(self.orig_vals))
        moly_inp = moly_inp + 'do_nothing t=2s\n'
        with open('moly.inp', 'w') as outfile:
            outfile.write(moly_inp)
        call('molywood moly.inp', shell=True)


def mappable(arg):
    """
    Wrapper for dual_annealing needed to make it parallelizable
    through Pool.map() in DihOpt.optimize() (needs to be defined in the
    global scope to be picklable)
    :param arg: tuple, combines system index, the DihOpt object, and the maxiter value
    :return: result of the dual_annealing fn
    """
    sys, dihopt, maxiter = arg
    return dual_annealing(dihopt.calc_rmse, args=(sys, True), bounds=dihopt.get_bounds(0), maxiter=maxiter, seed=sys,
                          x0=dihopt.get_opt_dih(dihopt.best_top), callback=dihopt.progress)

